# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/decks.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/decks.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import collection_pb2 as anki_dot_collection__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61nki/decks.proto\x12\nanki.decks\x1a\x12\x61nki/generic.proto\x1a\x15\x61nki/collection.proto\"\x15\n\x06\x44\x65\x63kId\x12\x0b\n\x03\x64id\x18\x01 \x01(\x03\"\x17\n\x07\x44\x65\x63kIds\x12\x0c\n\x04\x64ids\x18\x01 \x03(\x03\"\xd3\x0b\n\x04\x44\x65\x63k\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\nmtime_secs\x18\x03 \x01(\x03\x12\x0b\n\x03usn\x18\x04 \x01(\x05\x12\'\n\x06\x63ommon\x18\x05 \x01(\x0b\x32\x17.anki.decks.Deck.Common\x12)\n\x06normal\x18\x06 \x01(\x0b\x32\x17.anki.decks.Deck.NormalH\x00\x12-\n\x08\x66iltered\x18\x07 \x01(\x0b\x32\x19.anki.decks.Deck.FilteredH\x00\x1a\xd1\x01\n\x06\x43ommon\x12\x17\n\x0fstudy_collapsed\x18\x01 \x01(\x08\x12\x19\n\x11\x62rowser_collapsed\x18\x02 \x01(\x08\x12\x18\n\x10last_day_studied\x18\x03 \x01(\r\x12\x13\n\x0bnew_studied\x18\x04 \x01(\x05\x12\x16\n\x0ereview_studied\x18\x05 \x01(\x05\x12\x1c\n\x14milliseconds_studied\x18\x07 \x01(\x05\x12\x18\n\x10learning_studied\x18\x06 \x01(\x05\x12\x0e\n\x05other\x18\xff\x01 \x01(\x0cJ\x04\x08\x08\x10\x0e\x1a\xaa\x03\n\x06Normal\x12\x11\n\tconfig_id\x18\x01 \x01(\x03\x12\x12\n\nextend_new\x18\x02 \x01(\r\x12\x15\n\rextend_review\x18\x03 \x01(\r\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x1c\n\x14markdown_description\x18\x05 \x01(\x08\x12\x19\n\x0creview_limit\x18\x06 \x01(\rH\x00\x88\x01\x01\x12\x16\n\tnew_limit\x18\x07 \x01(\rH\x01\x88\x01\x01\x12<\n\x12review_limit_today\x18\x08 \x01(\x0b\x32 .anki.decks.Deck.Normal.DayLimit\x12\x39\n\x0fnew_limit_today\x18\t \x01(\x0b\x32 .anki.decks.Deck.Normal.DayLimit\x12\x1e\n\x11\x64\x65sired_retention\x18\n \x01(\x02H\x02\x88\x01\x01\x1a(\n\x08\x44\x61yLimit\x12\r\n\x05limit\x18\x01 \x01(\r\x12\r\n\x05today\x18\x02 \x01(\rB\x0f\n\r_review_limitB\x0c\n\n_new_limitB\x14\n\x12_desired_retentionJ\x04\x08\x0c\x10\x10\x1a\x90\x04\n\x08\x46iltered\x12\x12\n\nreschedule\x18\x01 \x01(\x08\x12:\n\x0csearch_terms\x18\x02 \x03(\x0b\x32$.anki.decks.Deck.Filtered.SearchTerm\x12\x0e\n\x06\x64\x65lays\x18\x03 \x03(\x02\x12\x15\n\rpreview_delay\x18\x04 \x01(\r\x12\x1a\n\x12preview_again_secs\x18\x07 \x01(\r\x12\x19\n\x11preview_hard_secs\x18\x05 \x01(\r\x12\x19\n\x11preview_good_secs\x18\x06 \x01(\r\x1a\xba\x02\n\nSearchTerm\x12\x0e\n\x06search\x18\x01 \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\r\x12\x39\n\x05order\x18\x03 \x01(\x0e\x32*.anki.decks.Deck.Filtered.SearchTerm.Order\"\xd1\x01\n\x05Order\x12\x19\n\x15OLDEST_REVIEWED_FIRST\x10\x00\x12\n\n\x06RANDOM\x10\x01\x12\x17\n\x13INTERVALS_ASCENDING\x10\x02\x12\x18\n\x14INTERVALS_DESCENDING\x10\x03\x12\n\n\x06LAPSES\x10\x04\x12\t\n\x05\x41\x44\x44\x45\x44\x10\x05\x12\x07\n\x03\x44UE\x10\x06\x12\x11\n\rREVERSE_ADDED\x10\x07\x12\x1c\n\x18RETRIEVABILITY_ASCENDING\x10\x08\x12\x1d\n\x19RETRIEVABILITY_DESCENDING\x10\t\x1aq\n\rKindContainer\x12)\n\x06normal\x18\x01 \x01(\x0b\x32\x17.anki.decks.Deck.NormalH\x00\x12-\n\x08\x66iltered\x18\x02 \x01(\x0b\x32\x19.anki.decks.Deck.FilteredH\x00\x42\x06\n\x04kindB\x06\n\x04kind\"L\n\x1c\x41\x64\x64OrUpdateDeckLegacyRequest\x12\x0c\n\x04\x64\x65\x63k\x18\x01 \x01(\x0c\x12\x1e\n\x16preserve_usn_and_mtime\x18\x02 \x01(\x08\"\x1e\n\x0f\x44\x65\x63kTreeRequest\x12\x0b\n\x03now\x18\x01 \x01(\x03\"\xf2\x02\n\x0c\x44\x65\x63kTreeNode\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05level\x18\x04 \x01(\r\x12\x11\n\tcollapsed\x18\x05 \x01(\x08\x12\x14\n\x0creview_count\x18\x06 \x01(\r\x12\x13\n\x0blearn_count\x18\x07 \x01(\r\x12\x11\n\tnew_count\x18\x08 \x01(\r\x12\x19\n\x11intraday_learning\x18\t \x01(\r\x12\"\n\x1ainterday_learning_uncapped\x18\n \x01(\r\x12\x14\n\x0cnew_uncapped\x18\x0b \x01(\r\x12\x17\n\x0freview_uncapped\x18\x0c \x01(\r\x12\x15\n\rtotal_in_deck\x18\r \x01(\r\x12 \n\x18total_including_children\x18\x0e \x01(\r\x12\x10\n\x08\x66iltered\x18\x10 \x01(\x08\x12*\n\x08\x63hildren\x18\x03 \x03(\x0b\x32\x18.anki.decks.DeckTreeNode\"\x9b\x01\n\x17SetDeckCollapsedRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x11\n\tcollapsed\x18\x02 \x01(\x08\x12\x38\n\x05scope\x18\x03 \x01(\x0e\x32).anki.decks.SetDeckCollapsedRequest.Scope\"\"\n\x05Scope\x12\x0c\n\x08REVIEWER\x10\x00\x12\x0b\n\x07\x42ROWSER\x10\x01\"K\n\x13GetDeckNamesRequest\x12\x1a\n\x12skip_empty_default\x18\x01 \x01(\x08\x12\x18\n\x10include_filtered\x18\x02 \x01(\x08\"4\n\tDeckNames\x12\'\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x16.anki.decks.DeckNameId\"&\n\nDeckNameId\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\"<\n\x14ReparentDecksRequest\x12\x10\n\x08\x64\x65\x63k_ids\x18\x01 \x03(\x03\x12\x12\n\nnew_parent\x18\x02 \x01(\x03\"6\n\x11RenameDeckRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x10\n\x08new_name\x18\x02 \x01(\t\"q\n\x15\x46ilteredDeckForUpdate\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12)\n\x06\x63onfig\x18\x03 \x01(\x0b\x32\x19.anki.decks.Deck.Filtered\x12\x13\n\x0b\x61llow_empty\x18\x04 \x01(\x08\x32\xfa\x0c\n\x0c\x44\x65\x63ksService\x12\x30\n\x07NewDeck\x12\x13.anki.generic.Empty\x1a\x10.anki.decks.Deck\x12=\n\x07\x41\x64\x64\x44\x65\x63k\x12\x10.anki.decks.Deck\x1a .anki.collection.OpChangesWithId\x12\x45\n\rAddDeckLegacy\x12\x12.anki.generic.Json\x1a .anki.collection.OpChangesWithId\x12U\n\x15\x41\x64\x64OrUpdateDeckLegacy\x12(.anki.decks.AddOrUpdateDeckLegacyRequest\x1a\x12.anki.decks.DeckId\x12\x41\n\x08\x44\x65\x63kTree\x12\x1b.anki.decks.DeckTreeRequest\x1a\x18.anki.decks.DeckTreeNode\x12\x39\n\x0e\x44\x65\x63kTreeLegacy\x12\x13.anki.generic.Empty\x1a\x12.anki.generic.Json\x12<\n\x11GetAllDecksLegacy\x12\x13.anki.generic.Empty\x1a\x12.anki.generic.Json\x12;\n\x0fGetDeckIdByName\x12\x14.anki.generic.String\x1a\x12.anki.decks.DeckId\x12/\n\x07GetDeck\x12\x12.anki.decks.DeckId\x1a\x10.anki.decks.Deck\x12:\n\nUpdateDeck\x12\x10.anki.decks.Deck\x1a\x1a.anki.collection.OpChanges\x12\x42\n\x10UpdateDeckLegacy\x12\x12.anki.generic.Json\x1a\x1a.anki.collection.OpChanges\x12S\n\x10SetDeckCollapsed\x12#.anki.decks.SetDeckCollapsedRequest\x1a\x1a.anki.collection.OpChanges\x12\x37\n\rGetDeckLegacy\x12\x12.anki.decks.DeckId\x1a\x12.anki.generic.Json\x12\x46\n\x0cGetDeckNames\x12\x1f.anki.decks.GetDeckNamesRequest\x1a\x15.anki.decks.DeckNames\x12\x41\n\x14GetDeckAndChildNames\x12\x12.anki.decks.DeckId\x1a\x15.anki.decks.DeckNames\x12\x37\n\rNewDeckLegacy\x12\x12.anki.generic.Bool\x1a\x12.anki.generic.Json\x12G\n\x0bRemoveDecks\x12\x13.anki.decks.DeckIds\x1a#.anki.collection.OpChangesWithCount\x12V\n\rReparentDecks\x12 .anki.decks.ReparentDecksRequest\x1a#.anki.collection.OpChangesWithCount\x12G\n\nRenameDeck\x12\x1d.anki.decks.RenameDeckRequest\x1a\x1a.anki.collection.OpChanges\x12P\n\x17GetOrCreateFilteredDeck\x12\x12.anki.decks.DeckId\x1a!.anki.decks.FilteredDeckForUpdate\x12^\n\x17\x41\x64\x64OrUpdateFilteredDeck\x12!.anki.decks.FilteredDeckForUpdate\x1a .anki.collection.OpChangesWithId\x12H\n\x17\x46ilteredDeckOrderLabels\x12\x13.anki.generic.Empty\x1a\x18.anki.generic.StringList\x12@\n\x0eSetCurrentDeck\x12\x12.anki.decks.DeckId\x1a\x1a.anki.collection.OpChanges\x12\x37\n\x0eGetCurrentDeck\x12\x13.anki.generic.Empty\x1a\x10.anki.decks.Deck2\x15\n\x13\x42\x61\x63kendDecksServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.decks_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_DECKID']._serialized_start=75
  _globals['_DECKID']._serialized_end=96
  _globals['_DECKIDS']._serialized_start=98
  _globals['_DECKIDS']._serialized_end=121
  _globals['_DECK']._serialized_start=124
  _globals['_DECK']._serialized_end=1615
  _globals['_DECK_COMMON']._serialized_start=323
  _globals['_DECK_COMMON']._serialized_end=532
  _globals['_DECK_NORMAL']._serialized_start=535
  _globals['_DECK_NORMAL']._serialized_end=961
  _globals['_DECK_NORMAL_DAYLIMIT']._serialized_start=862
  _globals['_DECK_NORMAL_DAYLIMIT']._serialized_end=902
  _globals['_DECK_FILTERED']._serialized_start=964
  _globals['_DECK_FILTERED']._serialized_end=1492
  _globals['_DECK_FILTERED_SEARCHTERM']._serialized_start=1178
  _globals['_DECK_FILTERED_SEARCHTERM']._serialized_end=1492
  _globals['_DECK_FILTERED_SEARCHTERM_ORDER']._serialized_start=1283
  _globals['_DECK_FILTERED_SEARCHTERM_ORDER']._serialized_end=1492
  _globals['_DECK_KINDCONTAINER']._serialized_start=1494
  _globals['_DECK_KINDCONTAINER']._serialized_end=1607
  _globals['_ADDORUPDATEDECKLEGACYREQUEST']._serialized_start=1617
  _globals['_ADDORUPDATEDECKLEGACYREQUEST']._serialized_end=1693
  _globals['_DECKTREEREQUEST']._serialized_start=1695
  _globals['_DECKTREEREQUEST']._serialized_end=1725
  _globals['_DECKTREENODE']._serialized_start=1728
  _globals['_DECKTREENODE']._serialized_end=2098
  _globals['_SETDECKCOLLAPSEDREQUEST']._serialized_start=2101
  _globals['_SETDECKCOLLAPSEDREQUEST']._serialized_end=2256
  _globals['_SETDECKCOLLAPSEDREQUEST_SCOPE']._serialized_start=2222
  _globals['_SETDECKCOLLAPSEDREQUEST_SCOPE']._serialized_end=2256
  _globals['_GETDECKNAMESREQUEST']._serialized_start=2258
  _globals['_GETDECKNAMESREQUEST']._serialized_end=2333
  _globals['_DECKNAMES']._serialized_start=2335
  _globals['_DECKNAMES']._serialized_end=2387
  _globals['_DECKNAMEID']._serialized_start=2389
  _globals['_DECKNAMEID']._serialized_end=2427
  _globals['_REPARENTDECKSREQUEST']._serialized_start=2429
  _globals['_REPARENTDECKSREQUEST']._serialized_end=2489
  _globals['_RENAMEDECKREQUEST']._serialized_start=2491
  _globals['_RENAMEDECKREQUEST']._serialized_end=2545
  _globals['_FILTEREDDECKFORUPDATE']._serialized_start=2547
  _globals['_FILTEREDDECKFORUPDATE']._serialized_end=2660
  _globals['_DECKSSERVICE']._serialized_start=2663
  _globals['_DECKSSERVICE']._serialized_end=4321
  _globals['_BACKENDDECKSSERVICE']._serialized_start=4323
  _globals['_BACKENDDECKSSERVICE']._serialized_end=4344
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/notes.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/notes.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import notetypes_pb2 as anki_dot_notetypes__pb2
from anki import collection_pb2 as anki_dot_collection__pb2
from anki import decks_pb2 as anki_dot_decks__pb2
from anki import cards_pb2 as anki_dot_cards__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61nki/notes.proto\x12\nanki.notes\x1a\x14\x61nki/notetypes.proto\x1a\x15\x61nki/collection.proto\x1a\x10\x61nki/decks.proto\x1a\x10\x61nki/cards.proto\"\x15\n\x06NoteId\x12\x0b\n\x03nid\x18\x01 \x01(\x03\"\x1b\n\x07NoteIds\x12\x10\n\x08note_ids\x18\x01 \x03(\x03\"t\n\x04Note\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04guid\x18\x02 \x01(\t\x12\x13\n\x0bnotetype_id\x18\x03 \x01(\x03\x12\x12\n\nmtime_secs\x18\x04 \x01(\r\x12\x0b\n\x03usn\x18\x05 \x01(\x05\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x0e\n\x06\x66ields\x18\x07 \x03(\t\"A\n\x0e\x41\x64\x64NoteRequest\x12\x1e\n\x04note\x18\x01 \x01(\x0b\x32\x10.anki.notes.Note\x12\x0f\n\x07\x64\x65\x63k_id\x18\x02 \x01(\x03\"X\n\x0f\x41\x64\x64NoteResponse\x12\x34\n\x07\x63hanges\x18\x01 \x01(\x0b\x32#.anki.collection.OpChangesWithCount\x12\x0f\n\x07note_id\x18\x02 \x01(\x03\"?\n\x0f\x41\x64\x64NotesRequest\x12,\n\x08requests\x18\x01 \x03(\x0b\x32\x1a.anki.notes.AddNoteRequest\"M\n\x10\x41\x64\x64NotesResponse\x12+\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1a.anki.collection.OpChanges\x12\x0c\n\x04nids\x18\x02 \x03(\x03\"N\n\x12UpdateNotesRequest\x12\x1f\n\x05notes\x18\x01 \x03(\x0b\x32\x10.anki.notes.Note\x12\x17\n\x0fskip_undo_entry\x18\x02 \x01(\x08\"D\n\x18\x44\x65\x66\x61ultsForAddingRequest\x12(\n home_deck_of_current_review_card\x18\x01 \x01(\x03\"7\n\x0f\x44\x65\x63kAndNotetype\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x13\n\x0bnotetype_id\x18\x02 \x01(\x03\"8\n\x12RemoveNotesRequest\x12\x10\n\x08note_ids\x18\x01 \x03(\x03\x12\x10\n\x08\x63\x61rd_ids\x18\x02 \x03(\x03\"-\n\x1a\x43lozeNumbersInNoteResponse\x12\x0f\n\x07numbers\x18\x01 \x03(\r\"\\\n\x17\x41\x66terNoteUpdatesRequest\x12\x0c\n\x04nids\x18\x01 \x03(\x03\x12\x1b\n\x13mark_notes_modified\x18\x02 \x01(\x08\x12\x16\n\x0egenerate_cards\x18\x03 \x01(\x08\")\n\x19\x46ieldNamesForNotesRequest\x12\x0c\n\x04nids\x18\x01 \x03(\x03\",\n\x1a\x46ieldNamesForNotesResponse\x12\x0e\n\x06\x66ields\x18\x01 \x03(\t\"\xc2\x01\n\x17NoteFieldsCheckResponse\x12\x38\n\x05state\x18\x01 \x01(\x0e\x32).anki.notes.NoteFieldsCheckResponse.State\"m\n\x05State\x12\n\n\x06NORMAL\x10\x00\x12\t\n\x05\x45MPTY\x10\x01\x12\r\n\tDUPLICATE\x10\x02\x12\x11\n\rMISSING_CLOZE\x10\x03\x12\x16\n\x12NOTETYPE_NOT_CLOZE\x10\x04\x12\x13\n\x0f\x46IELD_NOT_CLOZE\x10\x05\x32\xa6\x08\n\x0cNotesService\x12\x37\n\x07NewNote\x12\x1a.anki.notetypes.NotetypeId\x1a\x10.anki.notes.Note\x12\x42\n\x07\x41\x64\x64Note\x12\x1a.anki.notes.AddNoteRequest\x1a\x1b.anki.notes.AddNoteResponse\x12\x45\n\x08\x41\x64\x64Notes\x12\x1b.anki.notes.AddNotesRequest\x1a\x1c.anki.notes.AddNotesResponse\x12V\n\x11\x44\x65\x66\x61ultsForAdding\x12$.anki.notes.DefaultsForAddingRequest\x1a\x1b.anki.notes.DeckAndNotetype\x12H\n\x16\x44\x65\x66\x61ultDeckForNotetype\x12\x1a.anki.notetypes.NotetypeId\x1a\x12.anki.decks.DeckId\x12I\n\x0bUpdateNotes\x12\x1e.anki.notes.UpdateNotesRequest\x1a\x1a.anki.collection.OpChanges\x12/\n\x07GetNote\x12\x12.anki.notes.NoteId\x1a\x10.anki.notes.Note\x12R\n\x0bRemoveNotes\x12\x1e.anki.notes.RemoveNotesRequest\x1a#.anki.collection.OpChangesWithCount\x12N\n\x12\x43lozeNumbersInNote\x12\x10.anki.notes.Note\x1a&.anki.notes.ClozeNumbersInNoteResponse\x12\\\n\x10\x41\x66terNoteUpdates\x12#.anki.notes.AfterNoteUpdatesRequest\x1a#.anki.collection.OpChangesWithCount\x12\x63\n\x12\x46ieldNamesForNotes\x12%.anki.notes.FieldNamesForNotesRequest\x1a&.anki.notes.FieldNamesForNotesResponse\x12H\n\x0fNoteFieldsCheck\x12\x10.anki.notes.Note\x1a#.anki.notes.NoteFieldsCheckResponse\x12\x36\n\x0b\x43\x61rdsOfNote\x12\x12.anki.notes.NoteId\x1a\x13.anki.cards.CardIds\x12K\n\x18GetSingleNotetypeOfNotes\x12\x13.anki.notes.NoteIds\x1a\x1a.anki.notetypes.NotetypeId2\x15\n\x13\x42\x61\x63kendNotesServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.notes_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_NOTEID']._serialized_start=113
  _globals['_NOTEID']._serialized_end=134
  _globals['_NOTEIDS']._serialized_start=136
  _globals['_NOTEIDS']._serialized_end=163
  _globals['_NOTE']._serialized_start=165
  _globals['_NOTE']._serialized_end=281
  _globals['_ADDNOTEREQUEST']._serialized_start=283
  _globals['_ADDNOTEREQUEST']._serialized_end=348
  _globals['_ADDNOTERESPONSE']._serialized_start=350
  _globals['_ADDNOTERESPONSE']._serialized_end=438
  _globals['_ADDNOTESREQUEST']._serialized_start=440
  _globals['_ADDNOTESREQUEST']._serialized_end=503
  _globals['_ADDNOTESRESPONSE']._serialized_start=505
  _globals['_ADDNOTESRESPONSE']._serialized_end=582
  _globals['_UPDATENOTESREQUEST']._serialized_start=584
  _globals['_UPDATENOTESREQUEST']._serialized_end=662
  _globals['_DEFAULTSFORADDINGREQUEST']._serialized_start=664
  _globals['_DEFAULTSFORADDINGREQUEST']._serialized_end=732
  _globals['_DECKANDNOTETYPE']._serialized_start=734
  _globals['_DECKANDNOTETYPE']._serialized_end=789
  _globals['_REMOVENOTESREQUEST']._serialized_start=791
  _globals['_REMOVENOTESREQUEST']._serialized_end=847
  _globals['_CLOZENUMBERSINNOTERESPONSE']._serialized_start=849
  _globals['_CLOZENUMBERSINNOTERESPONSE']._serialized_end=894
  _globals['_AFTERNOTEUPDATESREQUEST']._serialized_start=896
  _globals['_AFTERNOTEUPDATESREQUEST']._serialized_end=988
  _globals['_FIELDNAMESFORNOTESREQUEST']._serialized_start=990
  _globals['_FIELDNAMESFORNOTESREQUEST']._serialized_end=1031
  _globals['_FIELDNAMESFORNOTESRESPONSE']._serialized_start=1033
  _globals['_FIELDNAMESFORNOTESRESPONSE']._serialized_end=1077
  _globals['_NOTEFIELDSCHECKRESPONSE']._serialized_start=1080
  _globals['_NOTEFIELDSCHECKRESPONSE']._serialized_end=1274
  _globals['_NOTEFIELDSCHECKRESPONSE_STATE']._serialized_start=1165
  _globals['_NOTEFIELDSCHECKRESPONSE_STATE']._serialized_end=1274
  _globals['_NOTESSERVICE']._serialized_start=1277
  _globals['_NOTESSERVICE']._serialized_end=2339
  _globals['_BACKENDNOTESSERVICE']._serialized_start=2341
  _globals['_BACKENDNOTESSERVICE']._serialized_end=2362
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

import enum
import inspect
import os
import re
import shutil
import subprocess
import sys
from collections.abc import Callable, Sequence
from functools import partial, wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Union

from send2trash import send2trash

import aqt
from anki._legacy import DeprecatedNamesMixinForModule
from anki.collection import Collection, HelpPage
from anki.lang import TR, tr_legacyglobal  # noqa: F401
from anki.utils import (
    call,
    invalid_filename,
    is_mac,
    is_win,
    no_bundled_libs,
    version_with_build,
)
from aqt.qt import *
from aqt.qt import (
    PYQT_VERSION_STR,
    QT_VERSION_STR,  # noqa: F401
    QAction,
    QApplication,
    QCheckBox,
    QColor,
    QComboBox,
    QDesktopServices,
    QDialog,
    QDialogButtonBox,
    QEvent,
    QFileDialog,
    QFrame,
    QHeaderView,
    QIcon,
    QKeySequence,
    QLabel,
    QLineEdit,
    QListWidget,
    QMainWindow,
    QMenu,
    QMessageBox,
    QMouseEvent,
    QNativeGestureEvent,
    QOffscreenSurface,
    QOpenGLContext,
    QPalette,
    QPixmap,
    QPlainTextEdit,
    QPoint,
    QPushButton,
    QShortcut,
    QSize,
    QSplitter,
    QStandardPaths,
    Qt,
    QTextBrowser,
    QTextOption,
    QTimer,
    QUrl,
    QVBoxLayout,
    QWheelEvent,
    QWidget,
    pyqtSlot,
    qconnect,
    qtmajor,
    qtminor,
    qVersion,
    traceback,
)
from aqt.theme import theme_manager

if TYPE_CHECKING:
    TextFormat = Literal["plain", "rich", "markdown"]


def aqt_data_path() -> Path:
    import _aqt.colors

    data_folder = Path(inspect.getfile(_aqt.colors)).with_name("data")
    if data_folder.exists():
        return data_folder.absolute()
    else:
        # should only happen when running unit tests
        print("warning, data folder not found")
        return Path(".")


def aqt_data_folder() -> str:
    return str(aqt_data_path())


# shortcut to access Fluent translations; set as
tr = tr_legacyglobal

HelpPageArgument = Union["HelpPage.V", str]


def openHelp(section: HelpPageArgument) -> None:
    assert tr.backend is not None
    backend = tr.backend()
    assert backend is not None
    if isinstance(section, str):
        link = backend.help_page_link(page=HelpPage.INDEX) + section
    else:
        link = backend.help_page_link(page=section)
    openLink(link)


def openLink(link: str | QUrl) -> None:
    tooltip(tr.qt_misc_loading(), period=1000)
    with no_bundled_libs():
        QDesktopServices.openUrl(QUrl(link))


class MessageBox(QMessageBox):
    def __init__(
        self,
        text: str,
        callback: Callable[[int], None] | None = None,
        parent: QWidget | None = None,
        icon: QMessageBox.Icon = QMessageBox.Icon.NoIcon,
        help: HelpPageArgument | None = None,
        title: str = "Anki",
        buttons: (
            Sequence[
                str | QMessageBox.StandardButton | tuple[str, QMessageBox.ButtonRole]
            ]
            | None
        ) = None,
        default_button: int = 0,
        textFormat: Qt.TextFormat = Qt.TextFormat.PlainText,
        modality: Qt.WindowModality = Qt.WindowModality.WindowModal,
    ) -> None:
        parent = parent or aqt.mw.app.activeWindow() or aqt.mw
        super().__init__(parent)
        self.setText(text)
        self.setWindowTitle(title)
        self.setWindowModality(modality)
        self.setIcon(icon)
        if icon == QMessageBox.Icon.Question and theme_manager.night_mode:
            img = self.iconPixmap().toImage()
            img.invertPixels()
            self.setIconPixmap(QPixmap(img))
        self.setTextFormat(textFormat)
        if buttons is None:
            buttons = [QMessageBox.StandardButton.Ok]
        for i, button in enumerate(buttons):
            if isinstance(button, str):
                b = self.addButton(button, QMessageBox.ButtonRole.ActionRole)
            elif isinstance(button, QMessageBox.StandardButton):
                b = self.addButton(button)
                # a translator has complained the default Qt translation is inappropriate, so we override it
                if button == QMessageBox.StandardButton.Discard:
                    assert b is not None
                    b.setText(tr.actions_discard())
            elif isinstance(button, tuple):
                b = self.addButton(button[0], button[1])
            else:
                continue
            if callback is not None:
                assert b is not None
                qconnect(b.clicked, partial(callback, i))
            if i == default_button:
                self.setDefaultButton(b)
        if help is not None:
            b = self.addButton(QMessageBox.StandardButton.Help)
            assert b is not None
            qconnect(b.clicked, lambda: openHelp(help))
        self.open()


def ask_user(
    text: str,
    callback: Callable[[bool], None],
    defaults_yes: bool = True,
    **kwargs: Any,
) -> MessageBox:
    "Shows a yes/no question, passes the answer to the callback function as a bool."
    return MessageBox(
        text,
        callback=lambda response: callback(not response),
        icon=QMessageBox.Icon.Question,
        buttons=[QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No],
        default_button=not defaults_yes,
        **kwargs,
    )


def ask_user_dialog(
    text: str,
    callback: Callable[[int], None],
    buttons: (
        Sequence[str | QMessageBox.StandardButton | tuple[str, QMessageBox.ButtonRole]]
        | None
    ) = None,
    default_button: int = 1,
    parent: QWidget | None = None,
    title: str = "Anki",
    **kwargs: Any,
) -> MessageBox:
    "Shows a question to the user, passes the index of the button clicked to the callback."
    if buttons is None:
        buttons = [QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No]
    return MessageBox(
        text,
        callback=callback,
        icon=QMessageBox.Icon.Question,
        buttons=buttons,
        default_button=default_button,
        parent=parent,
        title=title,
        **kwargs,
    )


def show_info(
    text: str,
    callback: Callable | None = None,
    parent: QWidget | None = None,
    **kwargs: Any,
) -> MessageBox:
    "Show a small info window with an OK button."
    if "icon" not in kwargs:
        kwargs["icon"] = QMessageBox.Icon.Information
    return MessageBox(
        text,
        callback=(lambda _: callback()) if callback is not None else None,
        parent=parent,
        **kwargs,
    )


def show_warning(
    text: str,
    callback: Callable | None = None,
    parent: QWidget | None = None,
    **kwargs: Any,
) -> MessageBox:
    "Show a small warning window with an OK button."
    return show_info(
        text, icon=QMessageBox.Icon.Warning, callback=callback, parent=parent, **kwargs
    )


def show_critical(
    text: str,
    callback: Callable | None = None,
    parent: QWidget | None = None,
    **kwargs: Any,
) -> MessageBox:
    "Show a small critical error window with an OK button."
    return show_info(
        text, icon=QMessageBox.Icon.Critical, callback=callback, parent=parent, **kwargs
    )


def showWarning(
    text: str,
    parent: QWidget | None = None,
    help: HelpPageArgument | None = None,
    title: str = "Anki",
    textFormat: TextFormat | None = None,
) -> int:
    "Show a small warning with an OK button."
    return showInfo(text, parent, help, "warning", title=title, textFormat=textFormat)


def showCritical(
    text: str,
    parent: QDialog | None = None,
    help: str = "",
    title: str = "Anki",
    textFormat: TextFormat | None = None,
) -> int:
    "Show a small critical error with an OK button."
    return showInfo(text, parent, help, "critical", title=title, textFormat=textFormat)


def showInfo(
    text: str,
    parent: QWidget | None = None,
    help: HelpPageArgument | None = None,
    type: str = "info",
    title: str = "Anki",
    textFormat: TextFormat | None = None,
    customBtns: list[QMessageBox.StandardButton] | None = None,
) -> int:
    "Show a small info window with an OK button."
    parent_widget: QWidget
    if parent is None:
        parent_widget = aqt.mw.app.activeWindow() or aqt.mw
    else:
        parent_widget = parent
    if type == "warning":
        icon = QMessageBox.Icon.Warning
    elif type == "critical":
        icon = QMessageBox.Icon.Critical
    else:
        icon = QMessageBox.Icon.Information
    mb = QMessageBox(parent_widget)
    if textFormat == "plain":
        mb.setTextFormat(Qt.TextFormat.PlainText)
    elif textFormat == "rich":
        mb.setTextFormat(Qt.TextFormat.RichText)
    elif textFormat == "markdown":
        mb.setTextFormat(Qt.TextFormat.MarkdownText)
    elif textFormat is not None:
        raise Exception("unexpected textFormat type")
    mb.setText(text)
    mb.setIcon(icon)
    mb.setWindowTitle(title)
    if customBtns:
        default = None
        for btn in customBtns:
            b = mb.addButton(btn)
            if not default:
                default = b
        mb.setDefaultButton(default)
    else:
        b = mb.addButton(QMessageBox.StandardButton.Ok)
        assert b is not None
        b.setDefault(True)
    if help is not None:
        b = mb.addButton(QMessageBox.StandardButton.Help)
        assert b is not None
        qconnect(b.clicked, lambda: openHelp(help))
        b.setAutoDefault(False)
    return mb.exec()


def showText(
    txt: str,
    parent: QWidget | None = None,
    type: str = "text",
    run: bool = True,
    geomKey: str | None = None,
    minWidth: int = 500,
    minHeight: int = 400,
    title: str = "Anki",
    copyBtn: bool = False,
    plain_text_edit: bool = False,
) -> tuple[QDialog, QDialogButtonBox] | None:
    if not parent:
        parent = aqt.mw.app.activeWindow() or aqt.mw
    diag = QDialog(parent)
    diag.setWindowTitle(title)
    disable_help_button(diag)
    layout = QVBoxLayout(diag)
    diag.setLayout(layout)
    text: QPlainTextEdit | QTextBrowser
    if plain_text_edit:
        # used by the importer
        text = QPlainTextEdit()
        text.setReadOnly(True)
        text.setWordWrapMode(QTextOption.WrapMode.NoWrap)
        text.setPlainText(txt)
    else:
        text = QTextBrowser()
        text.setOpenExternalLinks(True)
        if type == "text":
            text.setPlainText(txt)
        else:
            text.setHtml(txt)
    layout.addWidget(text)
    box = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
    layout.addWidget(box)
    if copyBtn:

        def onCopy() -> None:
            clipboard = QApplication.clipboard()
            assert clipboard is not None
            clipboard.setText(text.toPlainText())

        btn = QPushButton(tr.qt_misc_copy_to_clipboard())
        qconnect(btn.clicked, onCopy)
        box.addButton(btn, QDialogButtonBox.ButtonRole.ActionRole)

    def onReject() -> None:
        if geomKey:
            saveGeom(diag, geomKey)
        QDialog.reject(diag)

    qconnect(box.rejected, onReject)

    def onFinish() -> None:
        if geomKey:
            saveGeom(diag, geomKey)

    qconnect(box.accepted, onFinish)
    diag.setMinimumHeight(minHeight)
    diag.setMinimumWidth(minWidth)
    if geomKey:
        restoreGeom(diag, geomKey)
    if run:
        diag.exec()
        return None
    else:
        return diag, box


def askUser(
    text: str,
    parent: QWidget | None = None,
    help: HelpPageArgument | None = None,
    defaultno: bool = False,
    msgfunc: Callable | None = None,
    title: str = "Anki",
) -> bool:
    "Show a yes/no question. Return true if yes."
    if not parent:
        parent = aqt.mw.app.activeWindow()
    if not msgfunc:
        msgfunc = QMessageBox.question
    sb = QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
    if help:
        sb |= QMessageBox.StandardButton.Help
    while 1:
        if defaultno:
            default = QMessageBox.StandardButton.No
        else:
            default = QMessageBox.StandardButton.Yes
        r = msgfunc(parent, title, text, sb, default)
        if r == QMessageBox.StandardButton.Help:
            assert help is not None
            openHelp(help)
        else:
            break
    return r == QMessageBox.StandardButton.Yes


class ButtonedDialog(QMessageBox):
    def __init__(
        self,
        text: str,
        buttons: list[str],
        parent: QWidget | None = None,
        help: HelpPageArgument | None = None,
        title: str = "Anki",
    ):
        QMessageBox.__init__(self, parent)
        self._buttons: list[QPushButton | None] = []
        self.setWindowTitle(title)
        self.help = help
        self.setIcon(QMessageBox.Icon.Warning)
        self.setText(text)
        for b in buttons:
            self._buttons.append(self.addButton(b, QMessageBox.ButtonRole.AcceptRole))
        if help:
            self.addButton(tr.actions_help(), QMessageBox.ButtonRole.HelpRole)
            buttons.append(tr.actions_help())

    def run(self) -> str:
        self.exec()
        clicked_button = self.clickedButton()
        assert clicked_button is not None
        txt = clicked_button.text()
        if txt == "Help":
            # FIXME stop dialog closing?
            assert self.help is not None
            openHelp(self.help)
        # work around KDE 'helpfully' adding accelerators to button text of Qt apps
        return txt.replace("&", "")

    def setDefault(self, idx: int) -> None:
        self.setDefaultButton(self._buttons[idx])


def askUserDialog(
    text: str,
    buttons: list[str],
    parent: QWidget | None = None,
    help: HelpPageArgument | None = None,
    title: str = "Anki",
) -> ButtonedDialog:
    if not parent:
        parent = aqt.mw
    diag = ButtonedDialog(text, buttons, parent, help, title=title)
    return diag


class GetTextDialog(QDialog):
    def __init__(
        self,
        parent: QWidget | None,
        question: str,
        help: HelpPageArgument | None = None,
        edit: QLineEdit | None = None,
        default: str = "",
        title: str = "Anki",
        minWidth: int = 400,
    ) -> None:
        QDialog.__init__(self, parent)
        self.setWindowTitle(title)
        disable_help_button(self)
        self.question = question
        self.help = help
        self.qlabel = QLabel(question)
        self.setMinimumWidth(minWidth)
        v = QVBoxLayout()
        v.addWidget(self.qlabel)
        if not edit:
            edit = QLineEdit()
        self.l = edit
        if default:
            self.l.setText(default)
            self.l.selectAll()
        v.addWidget(self.l)
        buts = (
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        if help:
            buts |= QDialogButtonBox.StandardButton.Help
        b = QDialogButtonBox(buts)  # type: ignore
        v.addWidget(b)
        self.setLayout(v)
        ok_button = b.button(QDialogButtonBox.StandardButton.Ok)
        assert ok_button is not None
        qconnect(ok_button.clicked, self.accept)

        cancel_button = b.button(QDialogButtonBox.StandardButton.Cancel)
        assert cancel_button is not None
        qconnect(cancel_button.clicked, self.reject)

        if help:
            help_button = b.button(QDialogButtonBox.StandardButton.Help)
            assert help_button is not None
            qconnect(help_button.clicked, self.helpRequested)
        self.l.setFocus()

    def accept(self) -> None:
        return QDialog.accept(self)

    def reject(self) -> None:
        return QDialog.reject(self)

    def helpRequested(self) -> None:
        if self.help is not None:
            openHelp(self.help)


def getText(
    prompt: str,
    parent: QWidget | None = None,
    help: HelpPageArgument | None = None,
    edit: QLineEdit | None = None,
    default: str = "",
    title: str = "Anki",
    geomKey: str | None = None,
    **kwargs: Any,
) -> tuple[str, int]:
    "Returns (string, succeeded)."
    if not parent:
        parent = aqt.mw.app.activeWindow() or aqt.mw
    d = GetTextDialog(
        parent, prompt, help=help, edit=edit, default=default, title=title, **kwargs
    )
    add_close_shortcut(d)
    d.setWindowModality(Qt.WindowModality.WindowModal)
    if geomKey:
        restoreGeom(d, geomKey)
    ret = d.exec()
    if geomKey and ret:
        saveGeom(d, geomKey)
    return (str(d.l.text()), ret)


def getOnlyText(*args: Any, **kwargs: Any) -> str:
    (s, r) = getText(*args, **kwargs)
    if r:
        return s
    else:
        return ""


# fixme: these utilities could be combined into a single base class
# unused by Anki, but used by add-ons
def chooseList(
    prompt: str, choices: list[str], startrow: int = 0, parent: Any | None = None
) -> int:
    if not parent:
        parent = aqt.mw.app.activeWindow()
    d = QDialog(parent)
    disable_help_button(d)
    d.setWindowModality(Qt.WindowModality.WindowModal)
    l = QVBoxLayout()
    d.setLayout(l)
    t = QLabel(prompt)
    l.addWidget(t)
    c = QListWidget()
    c.addItems(choices)
    c.setCurrentRow(startrow)
    l.addWidget(c)
    bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
    qconnect(bb.accepted, d.accept)
    l.addWidget(bb)
    d.exec()
    return c.currentRow()


def getTag(
    parent: QWidget, deck: Collection, question: str, **kwargs: Any
) -> tuple[str, int]:
    from aqt.tagedit import TagEdit

    te = TagEdit(parent)
    te.setCol(deck)
    ret = getText(question, parent, edit=te, geomKey="getTag", **kwargs)
    te.hideCompleter()
    return ret


def disable_help_button(widget: QWidget) -> None:
    "Disable the help button in the window titlebar."
    widget.setWindowFlags(
        widget.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint
    )


def setWindowIcon(widget: QWidget) -> None:
    icon = QIcon()
    icon.addPixmap(QPixmap("icons:anki.png"), QIcon.Mode.Normal, QIcon.State.Off)
    widget.setWindowIcon(icon)


# File handling
######################################################################


def getFile(
    parent: QWidget,
    title: str,
    # single file returned unless multi=True
    cb: Callable[[str | Sequence[str]], None] | None,
    filter: str = "*",
    dir: str | None = None,
    key: str | None = None,
    multi: bool = False,  # controls whether a single or multiple files is returned
) -> Sequence[str] | str | None:
    "Ask the user for a file."
    if dir and key:
        raise Exception("expected dir or key")
    if not dir:
        assert aqt.mw.pm.profile is not None
        dirkey = f"{key}Directory"
        dir = aqt.mw.pm.profile.get(dirkey, "")
    else:
        dirkey = None
    d = QFileDialog(parent)
    mode = (
        QFileDialog.FileMode.ExistingFiles
        if multi
        else QFileDialog.FileMode.ExistingFile
    )
    d.setFileMode(mode)
    assert dir is not None
    if os.path.exists(dir):
        d.setDirectory(dir)
    d.setWindowTitle(title)
    d.setNameFilter(filter)
    ret = []

    def accept() -> None:
        files = list(d.selectedFiles())
        if dirkey:
            assert aqt.mw.pm.profile is not None
            dir = os.path.dirname(files[0])
            aqt.mw.pm.profile[dirkey] = dir
        result = files if multi else files[0]
        if cb:
            cb(result)
        ret.append(result)

    qconnect(d.accepted, accept)
    if key:
        restoreState(d, key)
    d.exec()
    if key:
        saveState(d, key)
    return ret[0] if ret else None


def running_in_sandbox():
    """Check whether running in Flatpak or Snap. When in such a sandbox, Qt
    will not report the true location of user-chosen files, but instead a
    temporary location from which the sandboxing software will copy the file to
    the user-chosen destination. Thus file renames are impossible and caching
    the reported file location is unhelpful."""
    in_flatpak = (
        QStandardPaths.locate(
            QStandardPaths.StandardLocation.RuntimeLocation,
            "flatpak-info",
        )
        != ""
    )
    in_snap = bool(os.environ.get("SNAP"))
    return in_flatpak or in_snap


def getSaveFile(
    parent: QDialog,
    title: str,
    dir_description: str,
    key: str,
    ext: str,
    fname: str = "",
) -> str | None:
    """Ask the user for a file to save. Use DIR_DESCRIPTION as config
    variable. The file dialog will default to open with FNAME."""
    assert aqt.mw.pm.profile is not None
    config_key = f"{dir_description}Directory"

    defaultPath = QStandardPaths.writableLocation(
        QStandardPaths.StandardLocation.DocumentsLocation
    )
    base = aqt.mw.pm.profile.get(config_key, defaultPath)
    path = os.path.join(base, fname)
    file = QFileDialog.getSaveFileName(
        parent,
        title,
        path,
        f"{key} (*{ext})",
        options=QFileDialog.Option.DontConfirmOverwrite,
    )[0]
    if file and not running_in_sandbox():
        # add extension
        if not file.lower().endswith(ext):
            file += ext
        # save new default
        dir = os.path.dirname(file)
        aqt.mw.pm.profile[config_key] = dir
        # check if it exists
        if os.path.exists(file) and not askUser(
            tr.qt_misc_this_file_exists_are_you_sure(), parent
        ):
            return None
    return file


class _QtStateKeyKind(enum.Enum):
    HEADER = enum.auto()
    SPLITTER = enum.auto()
    STATE = enum.auto()
    GEOMETRY = enum.auto()


def _qt_state_key(kind: _QtStateKeyKind, key: str) -> str:
    """Construct a key used to save/restore geometry, state, etc.

    Adds Qt version number to key so that different data is saved per Qt version,
    preventing crashes and bugs when restoring data saved with a different Qt version.
    """
    qt_suffix = f"{qtmajor}.{qtminor}" if qtmajor > 5 else ""
    return f"{key}{kind.name.capitalize()}{qt_suffix}"


def saveGeom(widget: QWidget, key: str) -> None:
    # restoring a fullscreen window breaks the tab functionality of 5.15
    if not widget.isFullScreen() or qtmajor == 6:
        assert aqt.mw.pm.profile is not None
        key = _qt_state_key(_QtStateKeyKind.GEOMETRY, key)
        aqt.mw.pm.profile[key] = widget.saveGeometry()


def restoreGeom(
    widget: QWidget,
    key: str,
    adjustSize: bool = False,
    default_size: tuple[int, int] | None = None,
) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.GEOMETRY, key)
    if existing_geom := aqt.mw.pm.profile.get(key):
        widget.restoreGeometry(existing_geom)
        ensureWidgetInScreenBoundaries(widget)
    elif adjustSize:
        widget.adjustSize()
    elif default_size:
        widget.resize(*default_size)


def ensureWidgetInScreenBoundaries(widget: QWidget) -> None:
    window = widget.window()
    assert window is not None
    handle = window.windowHandle()
    if not handle:
        # window has not yet been shown, retry later
        aqt.mw.progress.timer(
            50, lambda: ensureWidgetInScreenBoundaries(widget), False, parent=widget
        )
        return

    # ensure widget is smaller than screen bounds
    screen = handle.screen()
    assert screen is not None
    geom = screen.availableGeometry()
    wsize = widget.size()
    cappedWidth = min(geom.width(), wsize.width())
    cappedHeight = min(geom.height(), wsize.height())
    if cappedWidth > wsize.width() or cappedHeight > wsize.height():
        widget.resize(QSize(cappedWidth, cappedHeight))

    # ensure widget is inside top left
    wpos = widget.pos()
    x = max(geom.x(), wpos.x())
    y = max(geom.y(), wpos.y())
    # and bottom right
    x = min(x, geom.width() + geom.x() - cappedWidth)
    y = min(y, geom.height() + geom.y() - cappedHeight)
    if x != wpos.x() or y != wpos.y():
        widget.move(x, y)


def saveState(widget: QFileDialog | QMainWindow, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.STATE, key)
    aqt.mw.pm.profile[key] = widget.saveState()


def restoreState(widget: QFileDialog | QMainWindow, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.STATE, key)
    if data := aqt.mw.pm.profile.get(key):
        widget.restoreState(data)


def saveSplitter(widget: QSplitter, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.SPLITTER, key)
    aqt.mw.pm.profile[key] = widget.saveState()


def restoreSplitter(widget: QSplitter, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.SPLITTER, key)
    if data := aqt.mw.pm.profile.get(key):
        widget.restoreState(data)


def saveHeader(widget: QHeaderView, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.HEADER, key)
    aqt.mw.pm.profile[key] = widget.saveState()


def restoreHeader(widget: QHeaderView, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key = _qt_state_key(_QtStateKeyKind.HEADER, key)
    if state := aqt.mw.pm.profile.get(key):
        widget.restoreState(state)


def save_is_checked(widget: QCheckBox, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key += "IsChecked"
    aqt.mw.pm.profile[key] = widget.isChecked()


def restore_is_checked(widget: QCheckBox, key: str) -> None:
    assert aqt.mw.pm.profile is not None
    key += "IsChecked"
    if aqt.mw.pm.profile.get(key) is not None:
        widget.setChecked(aqt.mw.pm.profile[key])


def save_combo_index_for_session(widget: QComboBox, key: str) -> None:
    textKey = f"{key}ComboActiveText"
    indexKey = f"{key}ComboActiveIndex"
    aqt.mw.pm.session[textKey] = widget.currentText()
    aqt.mw.pm.session[indexKey] = widget.currentIndex()


def restore_combo_index_for_session(
    widget: QComboBox, history: list[str], key: str
) -> None:
    textKey = f"{key}ComboActiveText"
    indexKey = f"{key}ComboActiveIndex"
    text = aqt.mw.pm.session.get(textKey)
    index = aqt.mw.pm.session.get(indexKey)
    if text is not None and index is not None:
        if index < len(history) and history[index] == text:
            widget.setCurrentIndex(index)


def save_combo_history(comboBox: QComboBox, history: list[str], name: str) -> str:
    assert aqt.mw.pm.profile is not None
    name += "BoxHistory"
    line_edit = comboBox.lineEdit()
    assert line_edit is not None
    text_input = line_edit.text()
    if text_input in history:
        history.remove(text_input)
    history.insert(0, text_input)
    history = history[:50]
    comboBox.clear()
    comboBox.addItems(history)
    aqt.mw.pm.session[name] = text_input
    aqt.mw.pm.profile[name] = history
    return text_input


def restore_combo_history(comboBox: QComboBox, name: str) -> list[str]:
    assert aqt.mw.pm.profile is not None
    name += "BoxHistory"
    history = aqt.mw.pm.profile.get(name, [])
    comboBox.addItems([""] + history)
    if history:
        session_input = aqt.mw.pm.session.get(name)
        if session_input and session_input == history[0]:
            line_edit = comboBox.lineEdit()
            assert line_edit is not None
            line_edit.setText(session_input)
            line_edit.selectAll()
    return history


def mungeQA(col: Collection, txt: str) -> str:
    print("mungeQA() deprecated; use mw.prepare_card_text_for_display()")
    txt = col.media.escape_media_filenames(txt)
    return txt


def openFolder(path: str) -> None:
    if is_win:
        subprocess.run(["explorer", f"file://{path}"], check=False)
    else:
        with no_bundled_libs():
            QDesktopServices.openUrl(QUrl.fromLocalFile(path))


def show_in_folder(path: str) -> None:
    if is_win:
        _show_in_folder_win32(path)
    elif is_mac:
        script = f"""
        tell application "Finder"
            activate
            select POSIX file "{path}"
        end tell
        """
        call(osascript_to_args(script))
    else:
        # For linux, there are multiple file managers. Let's test if one of the
        # most common file managers is found and use it in case it is installed.
        # If none of this list are installed, use a fallback. The fallback
        # might open the image in a web browser, image viewer or others,
        # depending on the users defaults.
        file_managers = [
            "nautilus",  # GNOME
            "dolphin",  # KDE
            "pcmanfm",  # LXDE
            "thunar",  # XFCE
            "nemo",  # Cinnamon
            "caja",  # MATE
        ]

        available_file_manager = None

        # Test if a file manager is installed and use it, fallback otherwise
        for file_manager in file_managers:
            if shutil.which(file_manager):
                available_file_manager = file_manager
                break

        if available_file_manager:
            subprocess.run([available_file_manager, path], check=False)
        else:
            # Just open the file in any other platform
            with no_bundled_libs():
                QDesktopServices.openUrl(QUrl.fromLocalFile(path))


def _show_in_folder_win32(path: str) -> None:
    import win32con
    import win32gui

    from aqt import mw

    def focus_explorer():
        hwnd = win32gui.FindWindow("CabinetWClass", None)
        if hwnd:
            win32gui.ShowWindow(hwnd, win32con.SW_RESTORE)
            win32gui.SetForegroundWindow(hwnd)

    subprocess.run(["explorer", "/select,", path], check=False)
    mw.progress.single_shot(500, focus_explorer)


def osascript_to_args(script: str):
    args = [
        item
        for line in script.splitlines()
        for item in ("-e", line.strip())
        if line.strip()
    ]
    return ["osascript"] + args


def shortcut(key: str) -> str:
    if is_mac:
        return re.sub("(?i)ctrl", "Command", key)
    return key


def maybeHideClose(bbox: QDialogButtonBox) -> None:
    if is_mac:
        b = bbox.button(QDialogButtonBox.StandardButton.Close)
        if b:
            bbox.removeButton(b)


def add_close_shortcut(widg: QWidget) -> None:
    if not is_mac:
        return
    shortcut = QShortcut(QKeySequence("Ctrl+W"), widg)
    qconnect(shortcut.activated, widg.close)


def downArrow() -> str:
    if is_win:
        return "▼"
    # windows 10 is lacking the smaller arrow on English installs
    return "▾"


def current_window() -> QWidget | None:
    if widget := QApplication.focusWidget():
        return widget.window()
    else:
        return None


def send_to_trash(path: Path) -> None:
    "Place file/folder in recycling bin, or delete permanently on failure."
    if not path.exists():
        return
    try:
        send2trash(path)
    except Exception as exc:
        # Linux users may not have a trash folder set up
        print("trash failure:", path, exc)
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()


# Tooltips
######################################################################

_tooltipTimer: QTimer | None = None
_tooltipLabel: QLabel | None = None


def tooltip(
    msg: str,
    period: int = 3000,
    parent: QWidget | None = None,
    x_offset: int = 0,
    y_offset: int = 100,
) -> None:
    global _tooltipTimer, _tooltipLabel

    class CustomLabel(QLabel):
        silentlyClose = True

        def mousePressEvent(self, evt: QMouseEvent | None) -> None:
            assert evt is not None
            evt.accept()
            self.hide()

    closeTooltip()
    aw = parent or aqt.mw.app.activeWindow() or aqt.mw
    lab = CustomLabel(
        f"""<table cellpadding=10>
<tr>
<td>{msg}</td>
</tr>
</table>""",
        aw,
    )
    lab.setFrameStyle(QFrame.Shape.Panel)
    lab.setLineWidth(2)
    lab.setWindowFlags(Qt.WindowType.ToolTip)
    if not theme_manager.night_mode:
        p = QPalette()
        p.setColor(QPalette.ColorRole.Window, QColor("#feffc4"))
        p.setColor(QPalette.ColorRole.WindowText, QColor("#000000"))
        lab.setPalette(p)
    lab.move(aw.mapToGlobal(QPoint(0 + x_offset, aw.height() - y_offset)))
    lab.show()
    _tooltipTimer = aqt.mw.progress.timer(
        period, closeTooltip, False, requiresCollection=False, parent=aw
    )
    _tooltipLabel = lab


def closeTooltip() -> None:
    global _tooltipLabel, _tooltipTimer
    if _tooltipLabel:
        try:
            _tooltipLabel.deleteLater()
        except RuntimeError:
            # already deleted as parent window closed
            pass
        _tooltipLabel = None
    if _tooltipTimer:
        try:
            _tooltipTimer.deleteLater()
        except RuntimeError:
            pass
        _tooltipTimer = None


# true if invalid; print warning
def checkInvalidFilename(str: str, dirsep: bool = True) -> bool:
    bad = invalid_filename(str, dirsep)
    if bad:
        showWarning(tr.qt_misc_the_following_character_can_not_be(val=bad))
        return True
    return False


# Menus
######################################################################
# This code will be removed in the future, please don't rely on it.

MenuListChild = Union["SubMenu", QAction, "MenuItem", "MenuList"]


class MenuList:
    def __init__(self) -> None:
        traceback.print_stack(file=sys.stdout)
        print(
            "MenuList will be removed; please copy it into your add-on's code if you need it."
        )
        self.children: list[MenuListChild | None] = []

    def addItem(self, title: str, func: Callable) -> MenuItem:
        item = MenuItem(title, func)
        self.children.append(item)
        return item

    def addSeparator(self) -> None:
        self.children.append(None)

    def addMenu(self, title: str) -> SubMenu:
        submenu = SubMenu(title)
        self.children.append(submenu)
        return submenu

    def addChild(self, child: SubMenu | QAction | MenuList) -> None:
        self.children.append(child)

    def renderTo(self, qmenu: QMenu) -> None:
        for child in self.children:
            if child is None:
                qmenu.addSeparator()
            elif isinstance(child, QAction):
                qmenu.addAction(child)
            else:
                child.renderTo(qmenu)

    def popupOver(self, widget: QPushButton) -> None:
        qmenu = QMenu()
        self.renderTo(qmenu)
        qmenu.exec(widget.mapToGlobal(QPoint(0, 0)))


class SubMenu(MenuList):
    def __init__(self, title: str) -> None:
        super().__init__()
        self.title = title

    def renderTo(self, menu: QMenu) -> None:
        submenu = menu.addMenu(self.title)
        assert submenu is not None
        super().renderTo(submenu)


class MenuItem:
    def __init__(self, title: str, func: Callable) -> None:
        self.title = title
        self.func = func

    def renderTo(self, qmenu: QMenu) -> None:
        a = qmenu.addAction(self.title)
        assert a is not None
        qconnect(a.triggered, self.func)


def qtMenuShortcutWorkaround(qmenu: QMenu) -> None:
    for act in qmenu.actions():
        act.setShortcutVisibleInContextMenu(True)


######################################################################


def disallow_full_screen() -> bool:
    """Test for OpenGl on Windows, which is known to cause issues with full screen mode."""
    from aqt import mw
    from aqt.profiles import VideoDriver

    return is_win and (
        mw.pm.video_driver() == VideoDriver.OpenGL
        and not os.environ.get("ANKI_SOFTWAREOPENGL")
    )


def add_ellipsis_to_action_label(*actions: QAction | QPushButton) -> None:
    """Pass actions to add '...' to their labels, indicating that more input is
    required before they can be performed.

    This approach is used so that the same fluent translations can be used on
    mobile, where the '...' convention does not exist.
    """
    for action in actions:
        action.setText(tr.actions_with_ellipsis(action=action.text()))


def supportText() -> str:
    import platform

    from aqt import mw

    platname = platform.platform()

    return """\
Anki {} {}
Python {} Qt {} PyQt {}
Platform: {}
""".format(
        version_with_build(),
        "(ao)" if mw.addonManager.dirty else "",
        platform.python_version(),
        qVersion(),
        PYQT_VERSION_STR,
        platname,
    )


######################################################################


# adapted from version detection in qutebrowser
def opengl_vendor() -> str | None:
    if qtmajor != 5:
        return "unknown"
    old_context = QOpenGLContext.currentContext()
    old_surface = None if old_context is None else old_context.surface()

    surface = QOffscreenSurface()
    surface.create()

    ctx = QOpenGLContext()
    ok = ctx.create()
    if not ok:
        return None

    ok = ctx.makeCurrent(surface)
    if not ok:
        return None

    try:
        if ctx.isOpenGLES():
            # Can't use versionFunctions there
            return None

        vp = QOpenGLVersionProfile()  # type: ignore
        vp.setVersion(2, 0)

        try:
            vf = ctx.versionFunctions(vp)  # type: ignore
        except ImportError:
            return None

        if vf is None:
            return None

        return vf.glGetString(vf.GL_VENDOR)
    finally:
        ctx.doneCurrent()
        if old_context and old_surface:
            old_context.makeCurrent(old_surface)


def gfxDriverIsBroken() -> bool:
    driver = opengl_vendor()
    return driver == "nouveau"


######################################################################


def startup_info() -> Any:
    "Use subprocess.Popen(startupinfo=...) to avoid opening a console window."
    if sys.platform != "win32":
        return None
    si = subprocess.STARTUPINFO()  # pytype: disable=module-attr
    si.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # pytype: disable=module-attr
    return si


def ensure_editor_saved(func: Callable) -> Callable:
    """Ensure the current editor's note is saved before running the wrapped function.

    Must be used on functions that may be invoked from a shortcut key while the
    editor has focus. For functions that can't be activated while the editor has
    focus, you don't need this.

    Will look for the editor as self.editor.
    """

    @wraps(func)
    def decorated(self: Any, *args: Any, **kwargs: Any) -> None:
        self.editor.call_after_note_saved(lambda: func(self, *args, **kwargs))

    return decorated


def skip_if_selection_is_empty(func: Callable) -> Callable:
    """Make the wrapped method a no-op and show a hint if the table selection is empty."""

    @wraps(func)
    def decorated(self: Any, *args: Any, **kwargs: Any) -> None:
        if self.table.len_selection() > 0:
            func(self, *args, **kwargs)
        else:
            tooltip(tr.browsing_no_selection())

    return decorated


def no_arg_trigger(func: Callable) -> Callable:
    """Tells Qt this function takes no args.

    This ensures PyQt doesn't attempt to pass a `toggled` arg
    into functions connected to a `triggered` signal.
    """

    return pyqtSlot()(func)  # type: ignore


def is_gesture_or_zoom_event(evt: QEvent) -> bool:
    """If the event is a gesture and/or will trigger zoom.

    Includes zoom by pinching, and Ctrl-scrolling on Win and Linux.
    """

    return isinstance(evt, QNativeGestureEvent) or (
        isinstance(evt, QWheelEvent)
        and not is_mac
        and KeyboardModifiersPressed().control
    )


class KeyboardModifiersPressed:
    "Util for type-safe checks of currently-pressed modifier keys."

    def __init__(self) -> None:
        from aqt import mw

        self._modifiers = mw.app.keyboardModifiers()

    @property
    def shift(self) -> bool:
        return bool(self._modifiers & Qt.KeyboardModifier.ShiftModifier)

    @property
    def control(self) -> bool:
        return bool(self._modifiers & Qt.KeyboardModifier.ControlModifier)

    @property
    def alt(self) -> bool:
        return bool(self._modifiers & Qt.KeyboardModifier.AltModifier)

    @property
    def meta(self) -> bool:
        return bool(self._modifiers & Qt.KeyboardModifier.MetaModifier)


# add-ons attempting to import isMac from this module :-(
_deprecated_names = DeprecatedNamesMixinForModule(globals())


if not TYPE_CHECKING:

    def __getattr__(name: str) -> Any:
        return _deprecated_names.__getattr__(name)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from dataclasses import dataclass

from anki.collection import UndoStatus


@dataclass
class UndoActionsInfo:
    can_undo: bool
    can_redo: bool

    undo_text: str
    redo_text: str

    # menu item is hidden when legacy undo is active, since it can't be undone
    show_redo: bool

    @staticmethod
    def from_undo_status(status: UndoStatus) -> UndoActionsInfo:
        from aqt import tr

        return UndoActionsInfo(
            can_undo=bool(status.undo),
            can_redo=bool(status.redo),
            undo_text=(
                tr.undo_undo_action(val=status.undo) if status.undo else tr.undo_undo()
            ),
            redo_text=(
                tr.undo_redo_action(action=status.redo)
                if status.redo
                else tr.undo_redo()
            ),
            show_redo=status.last_step > 0,
        )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import os
import sys
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import TextIO, cast

import anki.cards
import aqt
import aqt.forms
from aqt import gui_hooks
from aqt.qt import *
from aqt.utils import (
    disable_help_button,
    restoreGeom,
    restoreSplitter,
    saveGeom,
    saveSplitter,
    send_to_trash,
    tr,
)


def show_debug_console() -> None:
    assert aqt.mw
    console = DebugConsole(aqt.mw)
    gui_hooks.debug_console_will_show(console)
    console.show()


SCRIPT_FOLDER = "debug_scripts"
UNSAVED_SCRIPT = "Unsaved script"


@dataclass
class Action:
    name: str
    shortcut: str
    action: Callable[[], None]


class DebugConsole(QDialog):
    silentlyClose = True
    _last_index = 0

    def __init__(self, parent: QWidget) -> None:
        self._buffers: dict[int, str] = {}
        super().__init__(parent)
        self._setup_ui()
        disable_help_button(self)
        restoreGeom(self, "DebugConsoleWindow")
        restoreSplitter(self.frm.splitter, "DebugConsoleWindow")

    def _setup_ui(self):
        self.frm = aqt.forms.debug.Ui_Dialog()
        self.frm.setupUi(self)
        self._text: QPlainTextEdit = self.frm.text
        self._log: QPlainTextEdit = self.frm.log
        self._script: QComboBox = self.frm.script
        self._setup_text_edits()
        self._setup_scripts()
        self._setup_actions()
        self._setup_context_menu()
        qconnect(self.frm.widgetsButton.clicked, self._on_widgetGallery)
        qconnect(self._script.currentIndexChanged, self._on_script_change)

    def _setup_text_edits(self):
        font = QFont("Consolas")
        if not font.exactMatch():
            font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        font.setPointSize(self._text.font().pointSize() + 1)
        self._text.setFont(font)
        self._log.setFont(font)

    def _setup_scripts(self) -> None:
        self._dir = Path(aqt.mw.pm.base).joinpath(SCRIPT_FOLDER)
        self._dir.mkdir(exist_ok=True)
        self._script.addItem(UNSAVED_SCRIPT)
        self._script.addItems(os.listdir(self._dir))

    def _setup_actions(self) -> None:
        for action in self._actions():
            qconnect(
                QShortcut(QKeySequence(action.shortcut), self).activated, action.action
            )

    def _actions(self):
        return [
            Action("Execute", "ctrl+return", self.onDebugRet),
            Action("Execute and print", "ctrl+shift+return", self.onDebugPrint),
            Action("Clear log", "ctrl+l", self._log.clear),
            Action("Clear code", "ctrl+shift+l", self._text.clear),
            Action("Save script", "ctrl+s", self._save_script),
            Action("Open script", "ctrl+o", self._open_script),
            Action("Delete script", "ctrl+d", self._delete_script),
        ]

    def reject(self) -> None:
        super().reject()
        saveSplitter(self.frm.splitter, "DebugConsoleWindow")
        saveGeom(self, "DebugConsoleWindow")

    def _on_script_change(self, new_index: int) -> None:
        self._buffers[self._last_index] = self._text.toPlainText()
        self._text.setPlainText(self._get_script(new_index) or "")
        self._last_index = new_index

    def _get_script(self, idx: int) -> str | None:
        if script := self._buffers.get(idx, ""):
            return script
        if path := self._get_item(idx):
            return path.read_text(encoding="utf8")
        return None

    def _get_item(self, idx: int) -> Path | None:
        if not idx:
            return None
        path = Path(self._script.itemText(idx))
        return path if path.is_absolute() else self._dir.joinpath(path)

    def _get_index(self, path: Path) -> int:
        return self._script.findText(self._path_to_item(path))

    def _path_to_item(self, path: Path) -> str:
        return path.name if path.is_relative_to(self._dir) else str(path)

    def _current_script_path(self) -> Path | None:
        return self._get_item(self._script.currentIndex())

    def _save_script(self) -> None:
        if not (path := self._current_script_path()):
            new_file = QFileDialog.getSaveFileName(
                self, directory=str(self._dir), filter="Python file (*.py)"
            )[0]
            if not new_file:
                return
            path = Path(new_file)

        path.write_text(self._text.toPlainText(), encoding="utf8")

        item = self._path_to_item(path)
        if (idx := self._get_index(path)) == -1:
            self._script.addItem(item)
            idx = self._script.count() - 1
        # update existing buffer, so text edit doesn't change when index changes
        self._buffers[idx] = self._text.toPlainText()
        self._script.setCurrentIndex(idx)

    def _open_script(self) -> None:
        file = QFileDialog.getOpenFileName(
            self, directory=str(self._dir), filter="Python file (*.py)"
        )[0]
        if not file:
            return

        path = Path(file)
        item = self._path_to_item(path)
        if (idx := self._get_index(path)) == -1:
            self._script.addItem(item)
            idx = self._script.count() - 1
        elif idx in self._buffers:
            del self._buffers[idx]

        if idx == self._script.currentIndex():
            self._text.setPlainText(path.read_text(encoding="utf8"))
        else:
            self._script.setCurrentIndex(idx)

    def _delete_script(self) -> None:
        if not (path := self._current_script_path()):
            return
        send_to_trash(path)
        deleted_idx = self._script.currentIndex()
        self._script.setCurrentIndex(0)
        self._script.removeItem(deleted_idx)
        self._drop_buffer_and_shift_keys(deleted_idx)

    def _drop_buffer_and_shift_keys(self, idx: int) -> None:
        def shift(old_idx: int) -> int:
            return old_idx - 1 if old_idx > idx else old_idx

        self._buffers = {shift(i): val for i, val in self._buffers.items() if i != idx}

    def _setup_context_menu(self) -> None:
        for text_edit in (self._log, self._text):
            text_edit.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            qconnect(
                text_edit.customContextMenuRequested,
                partial(self._on_context_menu, text_edit),
            )

    def _on_context_menu(self, text_edit: QPlainTextEdit) -> None:
        menu = text_edit.createStandardContextMenu()
        assert menu is not None
        menu.addSeparator()
        for action in self._actions():
            entry = menu.addAction(action.name)
            entry.setShortcut(QKeySequence(action.shortcut))
            qconnect(entry.triggered, action.action)
        menu.exec(QCursor.pos())

    def _on_widgetGallery(self) -> None:
        from aqt.widgetgallery import WidgetGallery

        self.widgetGallery = WidgetGallery(self)
        self.widgetGallery.show()

    def _captureOutput(self, on: bool) -> None:
        mw2 = self

        class Stream:
            def write(self, data: str) -> None:
                mw2._output += data

        if on:
            self._output = ""
            self._oldStderr = sys.stderr
            self._oldStdout = sys.stdout
            s = cast(TextIO, Stream())
            sys.stderr = s
            sys.stdout = s
        else:
            sys.stderr = self._oldStderr
            sys.stdout = self._oldStdout

    def _card_repr(self, card: anki.cards.Card | None) -> None:
        import copy
        import pprint

        if not card:
            print("no card")
            return

        print("Front:", card.question())
        print("\n")
        print("Back:", card.answer())

        print("\nNote:")
        note = copy.copy(card.note())
        for k, v in note.items():
            print(f"- {k}:", v)

        print("\n")
        del note.fields
        del note._fmap
        pprint.pprint(note.__dict__)

        print("\nCard:")
        c = copy.copy(card)
        c._render_output = None
        pprint.pprint(c.__dict__)

    def _debugCard(self) -> anki.cards.Card | None:
        assert aqt.mw
        card = aqt.mw.reviewer.card
        self._card_repr(card)
        return card

    def _debugBrowserCard(self) -> anki.cards.Card | None:
        card = aqt.dialogs._dialogs["Browser"][1].card
        self._card_repr(card)
        return card

    def onDebugPrint(self) -> None:
        cursor = self._text.textCursor()
        position = cursor.position()
        cursor.select(QTextCursor.SelectionType.LineUnderCursor)
        line = cursor.selectedText()
        whitespace, stripped = _split_off_leading_whitespace(line)
        pfx, sfx = "pp(", ")"
        if not stripped.startswith(pfx):
            line = f"{whitespace}{pfx}{stripped}{sfx}"
            cursor.insertText(line)
            cursor.setPosition(position + len(pfx))
            self._text.setTextCursor(cursor)
        self.onDebugRet()

    def onDebugRet(self) -> None:
        import pprint
        import traceback

        text = self._text.toPlainText()
        vars = {
            "card": self._debugCard,
            "bcard": self._debugBrowserCard,
            "mw": aqt.mw,
            "pp": pprint.pprint,
        }
        self._captureOutput(True)
        try:
            exec(text, vars)
        except Exception:
            self._output += traceback.format_exc()
        self._captureOutput(False)
        buf = ""
        for c, line in enumerate(text.strip().split("\n")):
            if c == 0:
                buf += f">>> {line}\n"
            else:
                buf += f"... {line}\n"
        try:
            to_append = buf + (self._output or "<no output>")
            to_append = gui_hooks.debug_console_did_evaluate_python(
                to_append, text, self.frm
            )
            self._log.appendPlainText(to_append)
        except UnicodeDecodeError:
            to_append = tr.qt_misc_non_unicode_text()
            to_append = gui_hooks.debug_console_did_evaluate_python(
                to_append, text, self.frm
            )
            self._log.appendPlainText(to_append)
        slider = self._log.verticalScrollBar()
        assert slider is not None
        slider.setValue(slider.maximum())
        self._log.ensureCursorVisible()


def _split_off_leading_whitespace(text: str) -> tuple[str, str]:
    stripped = text.lstrip()
    whitespace = text[: len(text) - len(stripped)]
    return whitespace, stripped



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

from anki.cards import Card
from anki.decks import DeckId
from anki.scheduler.legacy import SchedulerBaseWithLegacy


class DummyScheduler(SchedulerBaseWithLegacy):
    reps = 0

    def reset(self) -> None:
        pass

    def getCard(self) -> Card | None:
        raise Exception("v1/v2 scheduler no longer supported")

    def answerCard(self, card: Card, ease: int) -> None:
        raise Exception("v1/v2 scheduler no longer supported")

    def _is_finished(self) -> bool:
        return False

    @property
    def active_decks(self) -> list[DeckId]:
        return []

    def counts(self) -> tuple[int, int, int]:
        return (0, 0, 0)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import os
import re
import time
from concurrent.futures import Future

import aqt
import aqt.forms
import aqt.main
from anki import hooks
from anki.cards import CardId
from anki.decks import DeckId
from anki.exporting import Exporter, exporters
from aqt import gui_hooks
from aqt.errors import show_exception
from aqt.qt import *
from aqt.utils import (
    add_close_shortcut,
    checkInvalidFilename,
    disable_help_button,
    getSaveFile,
    showWarning,
    tooltip,
    tr,
)


class ExportDialog(QDialog):
    def __init__(
        self,
        mw: aqt.main.AnkiQt,
        did: DeckId | None = None,
        cids: list[CardId] | None = None,
        parent: QWidget | None = None,
    ):
        QDialog.__init__(self, parent or mw, Qt.WindowType.Window)
        self.mw = mw
        self.col = mw.col.weakref()
        self.frm = aqt.forms.exporting.Ui_ExportDialog()
        self.frm.setupUi(self)
        self.frm.legacy_support.setVisible(False)
        self.exporter: Exporter | None = None
        self.cids = cids
        disable_help_button(self)
        self.setup(did)
        add_close_shortcut(self)
        self.exec()

    def setup(self, did: DeckId | None) -> None:
        self.exporters = exporters(self.col)
        # if a deck specified, start with .apkg type selected
        idx = 0
        if did or self.cids:
            for c, (k, e) in enumerate(self.exporters):
                if e.ext == ".apkg":
                    idx = c
                    break
        self.frm.format.insertItems(0, [e[0] for e in self.exporters])
        self.frm.format.setCurrentIndex(idx)
        qconnect(self.frm.format.activated, self.exporterChanged)
        self.exporterChanged(idx)
        # deck list
        if self.cids is None:
            self.decks = [tr.exporting_all_decks()]
            self.decks.extend(d.name for d in self.col.decks.all_names_and_ids())
        else:
            self.decks = [tr.exporting_selected_notes()]
        self.frm.deck.addItems(self.decks)
        # save button
        b = QPushButton(tr.exporting_export())
        self.frm.buttonBox.addButton(b, QDialogButtonBox.ButtonRole.AcceptRole)
        # set default option if accessed through deck button
        if did:
            name = self.mw.col.decks.get(did)["name"]
            index = self.frm.deck.findText(name)
            self.frm.deck.setCurrentIndex(index)

    def exporterChanged(self, idx: int) -> None:
        self.exporter = self.exporters[idx][1](self.col)
        self.isApkg = self.exporter.ext == ".apkg"
        self.isVerbatim = getattr(self.exporter, "verbatim", False)
        self.isTextNote = getattr(self.exporter, "includeTags", False)
        self.frm.includeSched.setVisible(
            getattr(self.exporter, "includeSched", None) is not None
        )
        self.frm.includeMedia.setVisible(
            getattr(self.exporter, "includeMedia", None) is not None
        )
        self.frm.includeTags.setVisible(
            getattr(self.exporter, "includeTags", None) is not None
        )
        html = getattr(self.exporter, "includeHTML", None)
        if html is not None:
            self.frm.includeHTML.setVisible(True)
            self.frm.includeHTML.setChecked(html)
        else:
            self.frm.includeHTML.setVisible(False)
        # show deck list?
        self.frm.deck.setVisible(not self.isVerbatim)
        # used by the new export screen
        self.frm.includeDeck.setVisible(False)
        self.frm.includeNotetype.setVisible(False)
        self.frm.includeGuid.setVisible(False)

    def accept(self) -> None:
        self.exporter.includeSched = self.frm.includeSched.isChecked()
        self.exporter.includeMedia = self.frm.includeMedia.isChecked()
        self.exporter.includeTags = self.frm.includeTags.isChecked()
        self.exporter.includeHTML = self.frm.includeHTML.isChecked()
        idx = self.frm.deck.currentIndex()
        if self.cids is not None:
            # Browser Selection
            self.exporter.cids = self.cids
            self.exporter.did = None
        elif idx == 0:
            # All decks
            self.exporter.did = None
            self.exporter.cids = None
        else:
            # Deck idx-1 in the list of decks
            self.exporter.cids = None
            name = self.decks[self.frm.deck.currentIndex()]
            self.exporter.did = self.col.decks.id(name)
        if self.isVerbatim:
            name = time.strftime("-%Y-%m-%d@%H-%M-%S", time.localtime(time.time()))
            deck_name = tr.exporting_collection() + name
        else:
            # Get deck name and remove invalid filename characters
            deck_name = self.decks[self.frm.deck.currentIndex()]
            deck_name = re.sub('[\\\\/?<>:*|"^]', "_", deck_name)

        filename = f"{deck_name}{self.exporter.ext}"
        if callable(self.exporter.key):
            key_str = self.exporter.key(self.col)
        else:
            key_str = self.exporter.key
        while 1:
            file = getSaveFile(
                self,
                tr.actions_export(),
                "export",
                key_str,
                self.exporter.ext,
                fname=filename,
            )
            if not file:
                return
            if checkInvalidFilename(os.path.basename(file), dirsep=False):
                continue
            file = os.path.normpath(file)
            if os.path.commonprefix([self.mw.pm.base, file]) == self.mw.pm.base:
                showWarning("Please choose a different export location.")
                continue
            break
        self.hide()
        if file:
            # check we can write to file
            try:
                f = open(file, "wb")
                f.close()
            except OSError as e:
                showWarning(tr.exporting_couldnt_save_file(val=str(e)))
            else:
                os.unlink(file)

            # progress handler: old apkg exporter
            def exported_media_count(cnt: int) -> None:
                self.mw.taskman.run_on_main(
                    lambda: self.mw.progress.update(
                        label=tr.exporting_exported_media_file(count=cnt)
                    )
                )

            # progress handler: adaptor for new colpkg importer into old exporting screen.
            # don't rename this; there's a hack in pylib/exporting.py that assumes this
            # name
            def exported_media(progress: str) -> None:
                self.mw.taskman.run_on_main(
                    lambda: self.mw.progress.update(label=progress)
                )

            def do_export() -> None:
                self.exporter.exportInto(file)

            def on_done(future: Future) -> None:
                self.mw.progress.finish()
                hooks.media_files_did_export.remove(exported_media_count)
                hooks.legacy_export_progress.remove(exported_media)
                try:
                    # raises if exporter failed
                    future.result()
                except Exception as exc:
                    show_exception(parent=self.mw, exception=exc)
                    self.on_export_failed()
                else:
                    self.on_export_finished()

            gui_hooks.legacy_exporter_will_export(self.exporter)
            if self.isVerbatim:
                gui_hooks.collection_will_temporarily_close(self.mw.col)
            self.mw.progress.start()
            hooks.media_files_did_export.append(exported_media_count)
            hooks.legacy_export_progress.append(exported_media)

            self.mw.taskman.run_in_background(do_export, on_done)

    def on_export_finished(self) -> None:
        if self.isVerbatim:
            msg = tr.exporting_collection_exported()
            self.mw.reopen()
        elif self.isTextNote:
            msg = tr.exporting_note_exported(count=self.exporter.count)
        else:
            msg = tr.exporting_card_exported(count=self.exporter.count)
        gui_hooks.legacy_exporter_did_export(self.exporter)
        tooltip(msg, period=3000)
        QDialog.reject(self)

    def on_export_failed(self) -> None:
        if self.isVerbatim:
            self.mw.reopen()
        QDialog.reject(self)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/frontend.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/frontend.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import scheduler_pb2 as anki_dot_scheduler__pb2
from anki import generic_pb2 as anki_dot_generic__pb2
from anki import search_pb2 as anki_dot_search__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x61nki/frontend.proto\x12\ranki.frontend\x1a\x14\x61nki/scheduler.proto\x1a\x12\x61nki/generic.proto\x1a\x11\x61nki/search.proto\"\x83\x01\n\x1bSchedulingStatesWithContext\x12\x30\n\x06states\x18\x01 \x01(\x0b\x32 .anki.scheduler.SchedulingStates\x12\x32\n\x07\x63ontext\x18\x02 \x01(\x0b\x32!.anki.scheduler.SchedulingContext\"[\n\x1aSetSchedulingStatesRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x30\n\x06states\x18\x02 \x01(\x0b\x32 .anki.scheduler.SchedulingStates2\x86\x04\n\x0f\x46rontendService\x12\x61\n\x1eGetSchedulingStatesWithContext\x12\x13.anki.generic.Empty\x1a*.anki.frontend.SchedulingStatesWithContext\x12U\n\x13SetSchedulingStates\x12).anki.frontend.SetSchedulingStatesRequest\x1a\x13.anki.generic.Empty\x12\x36\n\nImportDone\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12?\n\x0fSearchInBrowser\x12\x17.anki.search.SearchNode\x1a\x13.anki.generic.Empty\x12\x43\n\x17\x64\x65\x63kOptionsRequireClose\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12<\n\x10\x64\x65\x63kOptionsReady\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12=\n\x11SaveCustomColours\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty2\x18\n\x16\x42\x61\x63kendFrontendServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.frontend_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_SCHEDULINGSTATESWITHCONTEXT']._serialized_start=100
  _globals['_SCHEDULINGSTATESWITHCONTEXT']._serialized_end=231
  _globals['_SETSCHEDULINGSTATESREQUEST']._serialized_start=233
  _globals['_SETSCHEDULINGSTATESREQUEST']._serialized_end=324
  _globals['_FRONTENDSERVICE']._serialized_start=327
  _globals['_FRONTENDSERVICE']._serialized_end=845
  _globals['_BACKENDFRONTENDSERVICE']._serialized_start=847
  _globals['_BACKENDFRONTENDSERVICE']._serialized_end=871
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import re
from collections.abc import Iterable

from anki.collection import Collection
from aqt import gui_hooks
from aqt.qt import *
from aqt.qt import sip


class TagEdit(QLineEdit):
    _completer: QCompleter | TagCompleter

    lostFocus = pyqtSignal()

    # 0 = tags, 1 = decks
    def __init__(self, parent: QWidget, type: int = 0) -> None:
        QLineEdit.__init__(self, parent)
        self.col: Collection | None = None
        self.model = QStringListModel()
        self.type = type
        if type == 0:
            self._completer = TagCompleter(self.model, parent, self)
        else:
            self._completer = QCompleter(self.model, parent)
        self._completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self._completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self._completer.setFilterMode(Qt.MatchFlag.MatchContains)
        self.setCompleter(self._completer)

    def setCol(self, col: Collection) -> None:
        "Set the current col, updating list of available tags."
        self.col = col
        l: Iterable[str]
        if self.type == 0:
            l = self.col.tags.all()
        else:
            l = (d.name for d in self.col.decks.all_names_and_ids())
        self.model.setStringList(l)

    def focusInEvent(self, evt: QFocusEvent | None) -> None:
        QLineEdit.focusInEvent(self, evt)

    def keyPressEvent(self, evt: QKeyEvent | None) -> None:
        assert evt is not None
        popup = self._completer.popup()
        assert popup is not None

        if evt.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down):
            # show completer on arrow key up/down
            if not popup.isVisible():
                self.showCompleter()
            return
        if (
            evt.key() == Qt.Key.Key_Tab
            and evt.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            # select next completion
            if not popup.isVisible():
                self.showCompleter()
            index = self._completer.currentIndex()
            popup.setCurrentIndex(index)
            cur_row = index.row()
            if not self._completer.setCurrentRow(cur_row + 1):
                self._completer.setCurrentRow(0)
            return
        if evt.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return) and popup.isVisible():
            # apply first completion if no suggestion selected
            selected_row = popup.currentIndex().row()
            if selected_row == -1:
                self._completer.setCurrentRow(0)
                index = self._completer.currentIndex()
                popup.setCurrentIndex(index)
            self.hideCompleter()
            QWidget.keyPressEvent(self, evt)
            return
        QLineEdit.keyPressEvent(self, evt)
        if not evt.text():
            # if it's a modifier, don't show
            return
        if evt.key() not in (
            Qt.Key.Key_Enter,
            Qt.Key.Key_Return,
            Qt.Key.Key_Escape,
            Qt.Key.Key_Space,
            Qt.Key.Key_Tab,
            Qt.Key.Key_Backspace,
            Qt.Key.Key_Delete,
        ):
            self.showCompleter()
        gui_hooks.tag_editor_did_process_key(self, evt)

    def showCompleter(self) -> None:
        self._completer.setCompletionPrefix(self.text())
        self._completer.complete()

    def focusOutEvent(self, evt: QFocusEvent | None) -> None:
        QLineEdit.focusOutEvent(self, evt)
        self.lostFocus.emit()  # type: ignore
        popup = self._completer.popup()
        assert popup is not None
        popup.hide()

    def hideCompleter(self) -> None:
        if sip.isdeleted(self._completer):  # type: ignore
            return
        popup = self._completer.popup()
        assert popup is not None
        popup.hide()


class TagCompleter(QCompleter):
    def __init__(
        self,
        model: QStringListModel,
        parent: QWidget,
        edit: TagEdit,
    ) -> None:
        QCompleter.__init__(self, model, parent)
        self.tags: list[str] = []
        self.edit = edit
        self.cursor: int | None = None

    def splitPath(self, tags: str | None) -> list[str]:
        assert tags is not None
        assert self.edit.col is not None
        stripped_tags = tags.strip()
        stripped_tags = re.sub("  +", " ", stripped_tags)
        self.tags = self.edit.col.tags.split(stripped_tags)
        self.tags.append("")
        p = self.edit.cursorPosition()
        if tags.endswith("  "):
            self.cursor = len(self.tags) - 1
        else:
            self.cursor = stripped_tags.count(" ", 0, p)
        return [self.tags[self.cursor]]

    def pathFromIndex(self, idx: QModelIndex) -> str:
        if self.cursor is None:
            return self.edit.text()
        ret = QCompleter.pathFromIndex(self, idx)
        self.tags[self.cursor] = ret
        try:
            self.tags.remove("")
        except ValueError:
            pass
        return f"{' '.join(self.tags)} "



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


"""
Tools for extending Anki.

A hook takes a function that does not return a value.

A filter takes a function that returns its first argument, optionally
modifying it.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import decorator

# You can find the definitions in ../tools/genhooks.py
from anki.hooks_gen import *

# Legacy hook handling
##############################################################################

_hooks: dict[str, list[Callable[..., Any]]] = {}


def runHook(hook: str, *args: Any) -> None:
    "Run all functions on hook."
    hookFuncs = _hooks.get(hook, None)
    if hookFuncs:
        for func in hookFuncs:
            try:
                func(*args)
            except Exception:
                hookFuncs.remove(func)
                raise


def runFilter(hook: str, arg: Any, *args: Any) -> Any:
    hookFuncs = _hooks.get(hook, None)
    if hookFuncs:
        for func in hookFuncs:
            try:
                arg = func(arg, *args)
            except Exception:
                hookFuncs.remove(func)
                raise
    return arg


def addHook(hook: str, func: Callable) -> None:
    "Add a function to hook. Ignore if already on hook."
    if not _hooks.get(hook, None):
        _hooks[hook] = []
    if func not in _hooks[hook]:
        _hooks[hook].append(func)


def remHook(hook: Any, func: Any) -> None:
    "Remove a function if is on hook."
    hook = _hooks.get(hook, [])
    if func in hook:
        hook.remove(func)


# Monkey patching
##############################################################################
# Please only use this for prototyping or for when hooks are not practical,
# as add-ons that use monkey patching are more likely to break when Anki is
# updated.
#
# If you call wrap() with pos='around', the original function will not be called
# automatically but can be called with _old().
def wrap(old: Any, new: Any, pos: str = "after") -> Callable:
    "Override an existing function."

    def repl(*args: Any, **kwargs: Any) -> Any:
        if pos == "after":
            old(*args, **kwargs)
            return new(*args, **kwargs)
        elif pos == "before":
            new(*args, **kwargs)
            return old(*args, **kwargs)
        else:
            return new(_old=old, *args, **kwargs)

    def decorator_wrapper(f: Any, *args: Any, **kwargs: Any) -> Any:
        return repl(*args, **kwargs)

    return decorator.decorator(decorator_wrapper)(old)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from markdown import markdown

from aqt.qt import QMessageBox, Qt, QUrl
from aqt.utils import MessageBox, getText, openLink, tr


def show_url_schemes_dialog() -> None:
    from aqt import mw

    default = " ".join(mw.pm.allowed_url_schemes())
    schemes, ok = getText(
        prompt=tr.preferences_url_scheme_prompt(),
        title=tr.preferences_url_schemes(),
        default=default,
    )
    if ok:
        mw.pm.set_allowed_url_schemes(schemes.split(" "))
        mw.pm.save()


def is_supported_scheme(url: QUrl) -> bool:
    from aqt import mw

    scheme = url.scheme().lower()
    allowed_schemes = mw.pm.allowed_url_schemes()

    return scheme in allowed_schemes or scheme in ["http", "https"]


def always_allow_scheme(url: QUrl) -> None:
    from aqt import mw

    scheme = url.scheme().lower()
    mw.pm.always_allow_scheme(scheme)


def open_url_if_supported_scheme(url: QUrl) -> None:
    from aqt import mw

    if is_supported_scheme(url):
        openLink(url)
    else:

        def on_button(idx: int) -> None:
            if idx == 0:
                openLink(url)
            elif idx == 1:
                always_allow_scheme(url)
                openLink(url)

        msg = markdown(
            tr.preferences_url_scheme_warning(link=url.toString(), scheme=url.scheme())
        )
        MessageBox(
            msg,
            buttons=[
                tr.preferences_url_scheme_allow_once(),
                tr.preferences_url_scheme_always_allow(),
                (tr.actions_cancel(), QMessageBox.ButtonRole.RejectRole),
            ],
            parent=mw,
            callback=on_button,
            textFormat=Qt.TextFormat.RichText,
            default_button=2,
            icon=QMessageBox.Icon.Warning,
        )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import anki.collection


class AnkiException(Exception):
    """
    General Anki exception that all custom exceptions raised by Anki should
    inherit from. Allows add-ons to easily identify Anki-native exceptions.

    When inheriting from a Python built-in exception other than `Exception`,
    please supply `AnkiException` as an additional inheritance:

    ```
    class MyNewAnkiException(ValueError, AnkiException):
        pass
    ```
    """


class BackendError(AnkiException):
    "An error originating from Anki's backend."

    def __init__(
        self,
        message: str,
        help_page: anki.collection.HelpPage.V | None,
        context: str | None,
        backtrace: str | None,
    ) -> None:
        super().__init__()
        self._message = message
        self.help_page = help_page
        self.context = context
        self.backtrace = backtrace

    def __str__(self) -> str:
        return self._message


class Interrupted(BackendError):
    pass


class NetworkError(BackendError):
    pass


class SyncErrorKind(Enum):
    AUTH = 1
    OTHER = 2


class SyncError(BackendError):
    def __init__(
        self,
        message: str,
        help_page: anki.collection.HelpPage.V | None,
        context: str | None,
        backtrace: str | None,
        kind: SyncErrorKind,
    ):
        self.kind = kind
        super().__init__(message, help_page, context, backtrace)


class BackendIOError(BackendError):
    pass


class CustomStudyError(BackendError):
    pass


class DBError(BackendError):
    pass


class CardTypeError(BackendError):
    pass


class TemplateError(BackendError):
    pass


class NotFoundError(BackendError):
    pass


class DeletedError(BackendError):
    pass


class ExistsError(BackendError):
    pass


class UndoEmpty(BackendError):
    pass


class FilteredDeckError(BackendError):
    pass


class InvalidInput(BackendError):
    pass


class SearchError(BackendError):
    pass


class SchedulerUpgradeRequired(BackendError):
    pass


class AbortSchemaModification(AnkiException):
    pass


# legacy
DeckRenameError = FilteredDeckError
AnkiError = AbortSchemaModification



# Form implementation generated from reading ui file 'qt/aqt/forms/models.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        Dialog.resize(396, 255)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.modelsList = QtWidgets.QListWidget(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.modelsList.sizePolicy().hasHeightForWidth())
        self.modelsList.setSizePolicy(sizePolicy)
        self.modelsList.setObjectName("modelsList")
        self.gridLayout_2.addWidget(self.modelsList, 1, 0, 1, 1)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(12)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Vertical)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Close|QtWidgets.QDialogButtonBox.StandardButton.Help)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_3.addWidget(self.buttonBox)
        self.gridLayout_2.addLayout(self.verticalLayout_3, 1, 1, 1, 1)
        self.verticalLayout_5.addLayout(self.gridLayout_2)
        spacerItem = QtWidgets.QSpacerItem(20, 6, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.verticalLayout_5.addItem(spacerItem)
        self.verticalLayout_5.setStretch(0, 100)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.notetypes_note_types())


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/collection.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/collection.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import sync_pb2 as anki_dot_sync__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61nki/collection.proto\x12\x0f\x61nki.collection\x1a\x12\x61nki/generic.proto\x1a\x0f\x61nki/sync.proto\"b\n\x15OpenCollectionRequest\x12\x17\n\x0f\x63ollection_path\x18\x01 \x01(\t\x12\x19\n\x11media_folder_path\x18\x02 \x01(\t\x12\x15\n\rmedia_db_path\x18\x03 \x01(\t\"7\n\x16\x43loseCollectionRequest\x12\x1d\n\x15\x64owngrade_to_schema11\x18\x01 \x01(\x08\")\n\x15\x43heckDatabaseResponse\x12\x10\n\x08problems\x18\x01 \x03(\t\"\xe1\x01\n\tOpChanges\x12\x0c\n\x04\x63\x61rd\x18\x01 \x01(\x08\x12\x0c\n\x04note\x18\x02 \x01(\x08\x12\x0c\n\x04\x64\x65\x63k\x18\x03 \x01(\x08\x12\x0b\n\x03tag\x18\x04 \x01(\x08\x12\x10\n\x08notetype\x18\x05 \x01(\x08\x12\x0e\n\x06\x63onfig\x18\x06 \x01(\x08\x12\x13\n\x0b\x64\x65\x63k_config\x18\x0b \x01(\x08\x12\r\n\x05mtime\x18\x0c \x01(\x08\x12\x15\n\rbrowser_table\x18\x07 \x01(\x08\x12\x17\n\x0f\x62rowser_sidebar\x18\x08 \x01(\x08\x12\x11\n\tnote_text\x18\t \x01(\x08\x12\x14\n\x0cstudy_queues\x18\n \x01(\x08\"<\n\rOpChangesOnly\x12+\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1a.anki.collection.OpChanges\"P\n\x12OpChangesWithCount\x12+\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1a.anki.collection.OpChanges\x12\r\n\x05\x63ount\x18\x02 \x01(\r\"J\n\x0fOpChangesWithId\x12+\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1a.anki.collection.OpChanges\x12\n\n\x02id\x18\x02 \x01(\x03\";\n\nUndoStatus\x12\x0c\n\x04undo\x18\x01 \x01(\t\x12\x0c\n\x04redo\x18\x02 \x01(\t\x12\x11\n\tlast_step\x18\x03 \x01(\r\"\xb5\x01\n\x12OpChangesAfterUndo\x12+\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1a.anki.collection.OpChanges\x12\x11\n\toperation\x18\x02 \x01(\t\x12\x1d\n\x15reverted_to_timestamp\x18\x03 \x01(\x03\x12/\n\nnew_status\x18\x04 \x01(\x0b\x32\x1b.anki.collection.UndoStatus\x12\x0f\n\x07\x63ounter\x18\x05 \x01(\r\"\xeb\x05\n\x08Progress\x12#\n\x04none\x18\x01 \x01(\x0b\x32\x13.anki.generic.EmptyH\x00\x12\x32\n\nmedia_sync\x18\x02 \x01(\x0b\x32\x1c.anki.sync.MediaSyncProgressH\x00\x12\x15\n\x0bmedia_check\x18\x03 \x01(\tH\x00\x12\x37\n\tfull_sync\x18\x04 \x01(\x0b\x32\".anki.collection.Progress.FullSyncH\x00\x12;\n\x0bnormal_sync\x18\x05 \x01(\x0b\x32$.anki.collection.Progress.NormalSyncH\x00\x12\x41\n\x0e\x64\x61tabase_check\x18\x06 \x01(\x0b\x32\'.anki.collection.Progress.DatabaseCheckH\x00\x12\x13\n\timporting\x18\x07 \x01(\tH\x00\x12\x13\n\texporting\x18\x08 \x01(\tH\x00\x12@\n\x0e\x63ompute_params\x18\t \x01(\x0b\x32&.anki.collection.ComputeParamsProgressH\x00\x12\x46\n\x11\x63ompute_retention\x18\n \x01(\x0b\x32).anki.collection.ComputeRetentionProgressH\x00\x12@\n\x0e\x63ompute_memory\x18\x0b \x01(\x0b\x32&.anki.collection.ComputeMemoryProgressH\x00\x1a.\n\x08\x46ullSync\x12\x13\n\x0btransferred\x18\x01 \x01(\r\x12\r\n\x05total\x18\x02 \x01(\r\x1a;\n\nNormalSync\x12\r\n\x05stage\x18\x01 \x01(\t\x12\r\n\x05\x61\x64\x64\x65\x64\x18\x02 \x01(\t\x12\x0f\n\x07removed\x18\x03 \x01(\t\x1aJ\n\rDatabaseCheck\x12\r\n\x05stage\x18\x01 \x01(\t\x12\x13\n\x0bstage_total\x18\x02 \x01(\r\x12\x15\n\rstage_current\x18\x03 \x01(\rB\x07\n\x05value\"w\n\x15\x43omputeParamsProgress\x12\x0f\n\x07\x63urrent\x18\x01 \x01(\r\x12\r\n\x05total\x18\x02 \x01(\r\x12\x0f\n\x07reviews\x18\x03 \x01(\r\x12\x16\n\x0e\x63urrent_preset\x18\x04 \x01(\r\x12\x15\n\rtotal_presets\x18\x05 \x01(\r\":\n\x18\x43omputeRetentionProgress\x12\x0f\n\x07\x63urrent\x18\x01 \x01(\r\x12\r\n\x05total\x18\x02 \x01(\r\"R\n\x15\x43omputeMemoryProgress\x12\x15\n\rcurrent_cards\x18\x01 \x01(\r\x12\x13\n\x0btotal_cards\x18\x02 \x01(\r\x12\r\n\x05label\x18\x03 \x01(\t\"X\n\x13\x43reateBackupRequest\x12\x15\n\rbackup_folder\x18\x01 \x01(\t\x12\r\n\x05\x66orce\x18\x02 \x01(\x08\x12\x1b\n\x13wait_for_completion\x18\x03 \x01(\x08\x32\xf7\x04\n\x11\x43ollectionService\x12L\n\rCheckDatabase\x12\x13.anki.generic.Empty\x1a&.anki.collection.CheckDatabaseResponse\x12\x41\n\rGetUndoStatus\x12\x13.anki.generic.Empty\x1a\x1b.anki.collection.UndoStatus\x12@\n\x04Undo\x12\x13.anki.generic.Empty\x1a#.anki.collection.OpChangesAfterUndo\x12@\n\x04Redo\x12\x13.anki.generic.Empty\x1a#.anki.collection.OpChangesAfterUndo\x12@\n\x12\x41\x64\x64\x43ustomUndoEntry\x12\x14.anki.generic.String\x1a\x14.anki.generic.UInt32\x12\x44\n\x10MergeUndoEntries\x12\x14.anki.generic.UInt32\x1a\x1a.anki.collection.OpChanges\x12@\n\x0eLatestProgress\x12\x13.anki.generic.Empty\x1a\x19.anki.collection.Progress\x12\x39\n\rSetWantsAbort\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12H\n\x16SetLoadBalancerEnabled\x12\x12.anki.generic.Bool\x1a\x1a.anki.collection.OpChanges2\xc4\x03\n\x18\x42\x61\x63kendCollectionService\x12M\n\x0eOpenCollection\x12&.anki.collection.OpenCollectionRequest\x1a\x13.anki.generic.Empty\x12O\n\x0f\x43loseCollection\x12\'.anki.collection.CloseCollectionRequest\x1a\x13.anki.generic.Empty\x12H\n\x0c\x43reateBackup\x12$.anki.collection.CreateBackupRequest\x1a\x12.anki.generic.Bool\x12\x41\n\x15\x41waitBackupCompletion\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12@\n\x0eLatestProgress\x12\x13.anki.generic.Empty\x1a\x19.anki.collection.Progress\x12\x39\n\rSetWantsAbort\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.EmptyB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.collection_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_OPENCOLLECTIONREQUEST']._serialized_start=79
  _globals['_OPENCOLLECTIONREQUEST']._serialized_end=177
  _globals['_CLOSECOLLECTIONREQUEST']._serialized_start=179
  _globals['_CLOSECOLLECTIONREQUEST']._serialized_end=234
  _globals['_CHECKDATABASERESPONSE']._serialized_start=236
  _globals['_CHECKDATABASERESPONSE']._serialized_end=277
  _globals['_OPCHANGES']._serialized_start=280
  _globals['_OPCHANGES']._serialized_end=505
  _globals['_OPCHANGESONLY']._serialized_start=507
  _globals['_OPCHANGESONLY']._serialized_end=567
  _globals['_OPCHANGESWITHCOUNT']._serialized_start=569
  _globals['_OPCHANGESWITHCOUNT']._serialized_end=649
  _globals['_OPCHANGESWITHID']._serialized_start=651
  _globals['_OPCHANGESWITHID']._serialized_end=725
  _globals['_UNDOSTATUS']._serialized_start=727
  _globals['_UNDOSTATUS']._serialized_end=786
  _globals['_OPCHANGESAFTERUNDO']._serialized_start=789
  _globals['_OPCHANGESAFTERUNDO']._serialized_end=970
  _globals['_PROGRESS']._serialized_start=973
  _globals['_PROGRESS']._serialized_end=1720
  _globals['_PROGRESS_FULLSYNC']._serialized_start=1528
  _globals['_PROGRESS_FULLSYNC']._serialized_end=1574
  _globals['_PROGRESS_NORMALSYNC']._serialized_start=1576
  _globals['_PROGRESS_NORMALSYNC']._serialized_end=1635
  _globals['_PROGRESS_DATABASECHECK']._serialized_start=1637
  _globals['_PROGRESS_DATABASECHECK']._serialized_end=1711
  _globals['_COMPUTEPARAMSPROGRESS']._serialized_start=1722
  _globals['_COMPUTEPARAMSPROGRESS']._serialized_end=1841
  _globals['_COMPUTERETENTIONPROGRESS']._serialized_start=1843
  _globals['_COMPUTERETENTIONPROGRESS']._serialized_end=1901
  _globals['_COMPUTEMEMORYPROGRESS']._serialized_start=1903
  _globals['_COMPUTEMEMORYPROGRESS']._serialized_end=1985
  _globals['_CREATEBACKUPREQUEST']._serialized_start=1987
  _globals['_CREATEBACKUPREQUEST']._serialized_end=2075
  _globals['_COLLECTIONSERVICE']._serialized_start=2078
  _globals['_COLLECTIONSERVICE']._serialized_end=2709
  _globals['_BACKENDCOLLECTIONSERVICE']._serialized_start=2712
  _globals['_BACKENDCOLLECTIONSERVICE']._serialized_end=3164
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/card_rendering.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/card_rendering.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import notes_pb2 as anki_dot_notes__pb2
from anki import notetypes_pb2 as anki_dot_notetypes__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x61nki/card_rendering.proto\x12\x13\x61nki.card_rendering\x1a\x12\x61nki/generic.proto\x1a\x10\x61nki/notes.proto\x1a\x14\x61nki/notetypes.proto\";\n\x14\x45xtractAvTagsRequest\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x15\n\rquestion_side\x18\x02 \x01(\x08\"R\n\x15\x45xtractAvTagsResponse\x12\x0c\n\x04text\x18\x01 \x01(\t\x12+\n\x07\x61v_tags\x18\x02 \x03(\x0b\x32\x1a.anki.card_rendering.AVTag\"V\n\x05\x41VTag\x12\x18\n\x0esound_or_video\x18\x01 \x01(\tH\x00\x12*\n\x03tts\x18\x02 \x01(\x0b\x32\x1b.anki.card_rendering.TTSTagH\x00\x42\x07\n\x05value\"]\n\x06TTSTag\x12\x12\n\nfield_text\x18\x01 \x01(\t\x12\x0c\n\x04lang\x18\x02 \x01(\t\x12\x0e\n\x06voices\x18\x03 \x03(\t\x12\r\n\x05speed\x18\x04 \x01(\x02\x12\x12\n\nother_args\x18\x05 \x03(\t\"G\n\x13\x45xtractLatexRequest\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x0b\n\x03svg\x18\x02 \x01(\x08\x12\x15\n\rexpand_clozes\x18\x03 \x01(\x08\"X\n\x14\x45xtractLatexResponse\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x32\n\x05latex\x18\x02 \x03(\x0b\x32#.anki.card_rendering.ExtractedLatex\"6\n\x0e\x45xtractedLatex\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x12\n\nlatex_body\x18\x02 \x01(\t\"\xbe\x01\n\x10\x45mptyCardsReport\x12\x0e\n\x06report\x18\x01 \x01(\t\x12G\n\x05notes\x18\x02 \x03(\x0b\x32\x38.anki.card_rendering.EmptyCardsReport.NoteWithEmptyCards\x1aQ\n\x12NoteWithEmptyCards\x12\x0f\n\x07note_id\x18\x01 \x01(\x03\x12\x10\n\x08\x63\x61rd_ids\x18\x02 \x03(\x03\x12\x18\n\x10will_delete_note\x18\x03 \x01(\x08\"U\n\x19RenderExistingCardRequest\x12\x0f\n\x07\x63\x61rd_id\x18\x01 \x01(\x03\x12\x0f\n\x07\x62rowser\x18\x02 \x01(\x08\x12\x16\n\x0epartial_render\x18\x03 \x01(\x08\"\xb1\x01\n\x1cRenderUncommittedCardRequest\x12\x1e\n\x04note\x18\x01 \x01(\x0b\x32\x10.anki.notes.Note\x12\x10\n\x08\x63\x61rd_ord\x18\x02 \x01(\r\x12\x33\n\x08template\x18\x03 \x01(\x0b\x32!.anki.notetypes.Notetype.Template\x12\x12\n\nfill_empty\x18\x04 \x01(\x08\x12\x16\n\x0epartial_render\x18\x05 \x01(\x08\"\x94\x01\n\"RenderUncommittedCardLegacyRequest\x12\x1e\n\x04note\x18\x01 \x01(\x0b\x32\x10.anki.notes.Note\x12\x10\n\x08\x63\x61rd_ord\x18\x02 \x01(\r\x12\x10\n\x08template\x18\x03 \x01(\x0c\x12\x12\n\nfill_empty\x18\x04 \x01(\x08\x12\x16\n\x0epartial_render\x18\x05 \x01(\x08\"\xca\x01\n\x12RenderCardResponse\x12\x41\n\x0equestion_nodes\x18\x01 \x03(\x0b\x32).anki.card_rendering.RenderedTemplateNode\x12?\n\x0c\x61nswer_nodes\x18\x02 \x03(\x0b\x32).anki.card_rendering.RenderedTemplateNode\x12\x0b\n\x03\x63ss\x18\x03 \x01(\t\x12\x11\n\tlatex_svg\x18\x04 \x01(\x08\x12\x10\n\x08is_empty\x18\x05 \x01(\x08\"x\n\x14RenderedTemplateNode\x12\x0e\n\x04text\x18\x01 \x01(\tH\x00\x12G\n\x0breplacement\x18\x02 \x01(\x0b\x32\x30.anki.card_rendering.RenderedTemplateReplacementH\x00\x42\x07\n\x05value\"X\n\x1bRenderedTemplateReplacement\x12\x12\n\nfield_name\x18\x01 \x01(\t\x12\x14\n\x0c\x63urrent_text\x18\x02 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x03 \x03(\t\";\n\x15RenderMarkdownRequest\x12\x10\n\x08markdown\x18\x01 \x01(\t\x12\x10\n\x08sanitize\x18\x02 \x01(\x08\"\x8c\x01\n\x10StripHtmlRequest\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x38\n\x04mode\x18\x02 \x01(\x0e\x32*.anki.card_rendering.StripHtmlRequest.Mode\"0\n\x04Mode\x12\n\n\x06NORMAL\x10\x00\x12\x1c\n\x18PRESERVE_MEDIA_FILENAMES\x10\x01\"G\n\x15HtmlToTextLineRequest\x12\x0c\n\x04text\x18\x01 \x01(\t\x12 \n\x18preserve_media_filenames\x18\x02 \x01(\x08\"M\n\x14\x43ompareAnswerRequest\x12\x10\n\x08\x65xpected\x18\x01 \x01(\t\x12\x10\n\x08provided\x18\x02 \x01(\t\x12\x11\n\tcombining\x18\x03 \x01(\x08\"=\n\x1c\x45xtractClozeForTypingRequest\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x0f\n\x07ordinal\x18\x02 \x01(\r\"\'\n\x13\x41llTtsVoicesRequest\x12\x10\n\x08validate\x18\x01 \x01(\x08\"\xb8\x01\n\x14\x41llTtsVoicesResponse\x12\x42\n\x06voices\x18\x01 \x03(\x0b\x32\x32.anki.card_rendering.AllTtsVoicesResponse.TtsVoice\x1a\\\n\x08TtsVoice\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x10\n\x08language\x18\x03 \x01(\t\x12\x16\n\tavailable\x18\x04 \x01(\x08H\x00\x88\x01\x01\x42\x0c\n\n_available\"T\n\x15WriteTtsStreamRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x10\n\x08voice_id\x18\x02 \x01(\t\x12\r\n\x05speed\x18\x03 \x01(\x02\x12\x0c\n\x04text\x18\x04 \x01(\t2\xf2\t\n\x14\x43\x61rdRenderingService\x12\x66\n\rExtractAvTags\x12).anki.card_rendering.ExtractAvTagsRequest\x1a*.anki.card_rendering.ExtractAvTagsResponse\x12\x63\n\x0c\x45xtractLatex\x12(.anki.card_rendering.ExtractLatexRequest\x1a).anki.card_rendering.ExtractLatexResponse\x12K\n\rGetEmptyCards\x12\x13.anki.generic.Empty\x1a%.anki.card_rendering.EmptyCardsReport\x12m\n\x12RenderExistingCard\x12..anki.card_rendering.RenderExistingCardRequest\x1a\'.anki.card_rendering.RenderCardResponse\x12s\n\x15RenderUncommittedCard\x12\x31.anki.card_rendering.RenderUncommittedCardRequest\x1a\'.anki.card_rendering.RenderCardResponse\x12\x7f\n\x1bRenderUncommittedCardLegacy\x12\x37.anki.card_rendering.RenderUncommittedCardLegacyRequest\x1a\'.anki.card_rendering.RenderCardResponse\x12\x39\n\x0bStripAvTags\x12\x14.anki.generic.String\x1a\x14.anki.generic.String\x12R\n\x0eRenderMarkdown\x12*.anki.card_rendering.RenderMarkdownRequest\x1a\x14.anki.generic.String\x12<\n\x0e\x45ncodeIriPaths\x12\x14.anki.generic.String\x1a\x14.anki.generic.String\x12<\n\x0e\x44\x65\x63odeIriPaths\x12\x14.anki.generic.String\x1a\x14.anki.generic.String\x12H\n\tStripHtml\x12%.anki.card_rendering.StripHtmlRequest\x1a\x14.anki.generic.String\x12R\n\x0eHtmlToTextLine\x12*.anki.card_rendering.HtmlToTextLineRequest\x1a\x14.anki.generic.String\x12P\n\rCompareAnswer\x12).anki.card_rendering.CompareAnswerRequest\x1a\x14.anki.generic.String\x12`\n\x15\x45xtractClozeForTyping\x12\x31.anki.card_rendering.ExtractClozeForTypingRequest\x1a\x14.anki.generic.String2\x9f\x02\n\x1b\x42\x61\x63kendCardRenderingService\x12H\n\tStripHtml\x12%.anki.card_rendering.StripHtmlRequest\x1a\x14.anki.generic.String\x12\x63\n\x0c\x41llTtsVoices\x12(.anki.card_rendering.AllTtsVoicesRequest\x1a).anki.card_rendering.AllTtsVoicesResponse\x12Q\n\x0eWriteTtsStream\x12*.anki.card_rendering.WriteTtsStreamRequest\x1a\x13.anki.generic.EmptyB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.card_rendering_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_EXTRACTAVTAGSREQUEST']._serialized_start=110
  _globals['_EXTRACTAVTAGSREQUEST']._serialized_end=169
  _globals['_EXTRACTAVTAGSRESPONSE']._serialized_start=171
  _globals['_EXTRACTAVTAGSRESPONSE']._serialized_end=253
  _globals['_AVTAG']._serialized_start=255
  _globals['_AVTAG']._serialized_end=341
  _globals['_TTSTAG']._serialized_start=343
  _globals['_TTSTAG']._serialized_end=436
  _globals['_EXTRACTLATEXREQUEST']._serialized_start=438
  _globals['_EXTRACTLATEXREQUEST']._serialized_end=509
  _globals['_EXTRACTLATEXRESPONSE']._serialized_start=511
  _globals['_EXTRACTLATEXRESPONSE']._serialized_end=599
  _globals['_EXTRACTEDLATEX']._serialized_start=601
  _globals['_EXTRACTEDLATEX']._serialized_end=655
  _globals['_EMPTYCARDSREPORT']._serialized_start=658
  _globals['_EMPTYCARDSREPORT']._serialized_end=848
  _globals['_EMPTYCARDSREPORT_NOTEWITHEMPTYCARDS']._serialized_start=767
  _globals['_EMPTYCARDSREPORT_NOTEWITHEMPTYCARDS']._serialized_end=848
  _globals['_RENDEREXISTINGCARDREQUEST']._serialized_start=850
  _globals['_RENDEREXISTINGCARDREQUEST']._serialized_end=935
  _globals['_RENDERUNCOMMITTEDCARDREQUEST']._serialized_start=938
  _globals['_RENDERUNCOMMITTEDCARDREQUEST']._serialized_end=1115
  _globals['_RENDERUNCOMMITTEDCARDLEGACYREQUEST']._serialized_start=1118
  _globals['_RENDERUNCOMMITTEDCARDLEGACYREQUEST']._serialized_end=1266
  _globals['_RENDERCARDRESPONSE']._serialized_start=1269
  _globals['_RENDERCARDRESPONSE']._serialized_end=1471
  _globals['_RENDEREDTEMPLATENODE']._serialized_start=1473
  _globals['_RENDEREDTEMPLATENODE']._serialized_end=1593
  _globals['_RENDEREDTEMPLATEREPLACEMENT']._serialized_start=1595
  _globals['_RENDEREDTEMPLATEREPLACEMENT']._serialized_end=1683
  _globals['_RENDERMARKDOWNREQUEST']._serialized_start=1685
  _globals['_RENDERMARKDOWNREQUEST']._serialized_end=1744
  _globals['_STRIPHTMLREQUEST']._serialized_start=1747
  _globals['_STRIPHTMLREQUEST']._serialized_end=1887
  _globals['_STRIPHTMLREQUEST_MODE']._serialized_start=1839
  _globals['_STRIPHTMLREQUEST_MODE']._serialized_end=1887
  _globals['_HTMLTOTEXTLINEREQUEST']._serialized_start=1889
  _globals['_HTMLTOTEXTLINEREQUEST']._serialized_end=1960
  _globals['_COMPAREANSWERREQUEST']._serialized_start=1962
  _globals['_COMPAREANSWERREQUEST']._serialized_end=2039
  _globals['_EXTRACTCLOZEFORTYPINGREQUEST']._serialized_start=2041
  _globals['_EXTRACTCLOZEFORTYPINGREQUEST']._serialized_end=2102
  _globals['_ALLTTSVOICESREQUEST']._serialized_start=2104
  _globals['_ALLTTSVOICESREQUEST']._serialized_end=2143
  _globals['_ALLTTSVOICESRESPONSE']._serialized_start=2146
  _globals['_ALLTTSVOICESRESPONSE']._serialized_end=2330
  _globals['_ALLTTSVOICESRESPONSE_TTSVOICE']._serialized_start=2238
  _globals['_ALLTTSVOICESRESPONSE_TTSVOICE']._serialized_end=2330
  _globals['_WRITETTSSTREAMREQUEST']._serialized_start=2332
  _globals['_WRITETTSSTREAMREQUEST']._serialized_end=2416
  _globals['_CARDRENDERINGSERVICE']._serialized_start=2419
  _globals['_CARDRENDERINGSERVICE']._serialized_end=3685
  _globals['_BACKENDCARDRENDERINGSERVICE']._serialized_start=3688
  _globals['_BACKENDCARDRENDERINGSERVICE']._serialized_end=3975
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""
Helper for running tasks on background threads.

See QueryOp() and CollectionOp() for higher-level routines.
"""

from __future__ import annotations

import traceback
from collections.abc import Callable
from concurrent.futures import Future
from concurrent.futures.thread import ThreadPoolExecutor
from threading import Lock, current_thread, main_thread
from typing import Any

import aqt
from anki.collection import Progress
from aqt.progress import ProgressUpdate
from aqt.qt import *

Closure = Callable[[], None]


class TaskManager(QObject):
    _closures_pending = pyqtSignal()

    def __init__(self, mw: aqt.AnkiQt) -> None:
        QObject.__init__(self)
        self.mw = mw.weakref()
        self._no_collection_executor = ThreadPoolExecutor()
        self._collection_executor = ThreadPoolExecutor(max_workers=1)
        self._closures: list[Closure] = []
        self._closures_lock = Lock()
        qconnect(self._closures_pending, self._on_closures_pending)

    def run_on_main(self, closure: Closure) -> None:
        "Run the provided closure on the main thread."
        with self._closures_lock:
            self._closures.append(closure)
        self._closures_pending.emit()  # type: ignore

    def run_in_background(
        self,
        task: Callable,
        on_done: Callable[[Future], None] | None = None,
        args: dict[str, Any] | None = None,
        uses_collection=True,
    ) -> Future:
        """Use QueryOp()/CollectionOp() in new code.

        Run task on a background thread.

        If on_done is provided, it will be called on the main thread with
        the completed future.

        Args if provided will be passed on as keyword arguments to the task callable.

        Tasks that access the collection are serialized. If you're doing things that
        don't require the collection (e.g. network requests), you can pass uses_collection
        =False to allow multiple tasks to run in parallel."""
        # Before we launch a background task, ensure any pending on_done closure are run on
        # main. Qt's signal/slot system will have posted a notification, but it may
        # not have been processed yet. The on_done() closures may make small queries
        # to the database that we want to run first - if we delay them until after the
        # background task starts, and it takes out a long-running lock on the database,
        # the UI thread will hang until the end of the op.
        if current_thread() is main_thread():
            self._on_closures_pending()
        else:
            print("bug: run_in_background not called from main thread")
            traceback.print_stack()

        if args is None:
            args = {}

        executor = (
            self._collection_executor
            if uses_collection
            else self._no_collection_executor
        )
        fut = executor.submit(task, **args)

        if on_done is not None:
            fut.add_done_callback(
                lambda future: self.run_on_main(lambda: on_done(future))
            )

        return fut

    def with_progress(
        self,
        task: Callable,
        on_done: Callable[[Future], None] | None = None,
        parent: QWidget | None = None,
        label: str | None = None,
        immediate: bool = False,
        uses_collection=True,
        title: str = "Anki",
    ) -> None:
        "Use QueryOp()/CollectionOp() in new code."
        self.mw.progress.start(
            parent=parent, label=label, immediate=immediate, title=title
        )

        def wrapped_done(fut: Future) -> None:
            self.mw.progress.finish()
            if on_done:
                on_done(fut)

        self.run_in_background(task, wrapped_done, uses_collection=uses_collection)

    def with_backend_progress(
        self,
        task: Callable,
        progress_update: Callable[[Progress, ProgressUpdate], None],
        on_done: Callable[[Future], None] | None = None,
        parent: QWidget | None = None,
        start_label: str | None = None,
        uses_collection=True,
    ) -> None:
        self.mw.progress.start_with_backend_updates(
            progress_update,
            parent=parent,
            start_label=start_label,
        )

        def wrapped_done(fut: Future) -> None:
            self.mw.progress.finish()
            # allow the event loop to close the window before we proceed
            if on_done:
                self.mw.progress.single_shot(
                    100, lambda: on_done(fut), requires_collection=False
                )

        self.run_in_background(task, wrapped_done, uses_collection=uses_collection)

    def _on_closures_pending(self) -> None:
        """Run any pending closures. This runs in the main thread."""
        with self._closures_lock:
            closures = self._closures
            self._closures = []

        for closure in closures:
            try:
                closure()
            except Exception as e:

                def raise_exception(exception=e) -> None:
                    raise exception

                QTimer.singleShot(0, raise_exception)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""Helpers for the packaged version of Anki."""

from __future__ import annotations

import contextlib
import os
import subprocess
import sys
from pathlib import Path

from anki.utils import is_mac, is_win


# ruff: noqa: F401
def first_run_setup() -> None:
    """Code run the first time after install/upgrade.

    Currently, we just import our main libraries and invoke
    mpv/lame on macOS, which is slow on the first run, and doing
    it this way shows progress being made.
    """

    if not is_mac:
        return

    # Import anki_audio first and spawn commands
    import anki_audio

    audio_pkg_path = Path(anki_audio.__file__).parent

    # Start mpv and lame commands concurrently
    processes = []
    for cmd_name in ["mpv", "lame"]:
        cmd_path = audio_pkg_path / cmd_name
        proc = subprocess.Popen(
            [str(cmd_path), "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        processes.append(proc)

    # Continue with other imports while commands run
    import concurrent.futures

    import bs4
    import flask
    import markdown
    import PyQt6.QtCore
    import PyQt6.QtGui
    import PyQt6.QtNetwork
    import PyQt6.QtQuick
    import PyQt6.QtWebChannel
    import PyQt6.QtWebEngineCore
    import PyQt6.QtWebEngineWidgets
    import PyQt6.QtWidgets
    import PyQt6.sip
    import requests
    import waitress

    import anki.collection

    from . import _macos_helper

    # Wait for both commands to complete
    for proc in processes:
        proc.wait()


def uv_binary() -> str | None:
    """Return the path to the uv binary."""
    return os.environ.get("ANKI_LAUNCHER_UV")


def launcher_root() -> str | None:
    """Return the path to the launcher root directory (AnkiProgramFiles)."""
    return os.environ.get("UV_PROJECT")


def venv_binary(cmd: str) -> str | None:
    """Return the path to a binary in the launcher's venv."""
    root = launcher_root()
    if not root:
        return None

    root_path = Path(root)
    if is_win:
        binary_path = root_path / ".venv" / "Scripts" / cmd
    else:
        binary_path = root_path / ".venv" / "bin" / cmd

    return str(binary_path)


def add_python_requirements(reqs: list[str]) -> tuple[bool, str]:
    """Add Python requirements to the launcher venv using uv add.

    Returns (success, output)"""

    binary = uv_binary()
    if not binary:
        return (False, "Not in packaged build.")

    uv_cmd = [binary, "add"] + reqs
    result = subprocess.run(uv_cmd, capture_output=True, text=True, check=False)

    if result.returncode == 0:
        root = launcher_root()
        if root:
            sync_marker = Path(root) / ".sync_complete"
            sync_marker.touch()

        return (True, result.stdout)
    else:
        return (False, result.stderr)


def launcher_executable() -> str | None:
    """Return the path to the Anki launcher executable."""
    return os.getenv("ANKI_LAUNCHER")


def trigger_launcher_run() -> None:
    """Create a trigger file to request launcher UI on next run."""
    try:
        root = launcher_root()
        if not root:
            return

        trigger_path = Path(root) / ".want-launcher"
        trigger_path.touch()
    except Exception as e:
        print(e)


def update_and_restart() -> None:
    """Update and restart Anki using the launcher."""
    from aqt import mw

    launcher = launcher_executable()
    assert launcher

    trigger_launcher_run()

    with contextlib.suppress(ResourceWarning):
        env = os.environ.copy()
        env["ANKI_LAUNCHER_WANT_TERMINAL"] = "1"
        # fixes a bug where launcher fails to appear if opening it
        # straight after updating
        if "GNOME_TERMINAL_SCREEN" in env:
            del env["GNOME_TERMINAL_SCREEN"]
        creationflags = 0
        if sys.platform == "win32":
            creationflags = (
                subprocess.CREATE_NEW_PROCESS_GROUP | subprocess.DETACHED_PROCESS
            )
        # On Windows 10, changing the handles breaks ANSI display
        io = None if sys.platform == "win32" else subprocess.DEVNULL

        subprocess.Popen(
            [launcher],
            start_new_session=True,
            stdin=io,
            stdout=io,
            stderr=io,
            env=env,
            creationflags=creationflags,
        )

    mw.app.quit()
